# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Test module for App, base class for all applications.
"""

import collections
from unittest.mock import Mock, call, patch

import pytest

from plinth import log
from plinth.app import (App, Component, EnableState, FollowerComponent, Info,
                        LeaderComponent, apps_init)
from plinth.diagnostic_check import DiagnosticCheck, Result

# pylint: disable=protected-access

privileged_modules_to_mock = ['plinth.privileged']


class AppTest(App):
    """Sample App for testing."""
    app_id = 'test-app'


class AppSetupTest(App):
    """Sample App for testing setup operations."""
    app_id = 'test-app-setup'

    def __init__(self):
        super().__init__()
        info = Info('test-app-setup', 3)
        self.add(info)


class LeaderTest(FollowerComponent, log.LogEmitter):
    """Test class for using LeaderComponent in tests."""
    is_leader = True

    @property
    def unit(self):  # For LogEmitter
        return self.component_id

    def diagnose(self) -> list[DiagnosticCheck]:
        """Return diagnostic results."""
        return [
            DiagnosticCheck('test-result-' + self.component_id,
                            'test-result-' + self.component_id, Result.PASSED)
        ]


@pytest.fixture(name='app_with_components')
def fixture_app_with_components():
    """Setup an app with some components."""
    app = AppTest()
    app.add(FollowerComponent('test-follower-1'))
    app.add(FollowerComponent('test-follower-2'))
    app.add(LeaderTest('test-leader-1'))
    app.add(LeaderTest('test-leader-2'))
    return app


@pytest.fixture(name='empty_apps', autouse=True)
def fixture_empty_apps():
    """Remove all apps from global list before starting a test."""
    App._all_apps = collections.OrderedDict()


def test_app_instantiation():
    """Test that App is instantiated properly."""
    app = AppTest()
    assert isinstance(app.components, collections.OrderedDict)
    assert not app.components
    assert app.app_id == 'test-app'
    assert app._all_apps['test-app'] == app
    assert len(app._all_apps) == 1
    assert app.can_be_disabled
    assert not app.locked
    assert app.configure_when_disabled


def test_app_get():
    """Test that an app can be correctly retrieved."""
    app = AppTest()
    assert App.get(app.app_id) == app


def test_app_list():
    """Test listing all apps."""
    app = AppTest()
    assert list(App.list()) == [app]


def test_app_add():
    """Test adding a components to an App."""
    app = AppTest()
    component = Component('test-component')
    return_value = app.add(component)
    assert len(app.components) == 1
    assert app.components['test-component'] == component
    assert component.app_id == app.app_id
    assert return_value == app


def test_app_remove(app_with_components):
    """Test removing a component from an App."""
    app = app_with_components
    component = app.components['test-leader-1']
    assert app.remove('test-leader-1') == component
    assert component.app_id is None
    assert 'test-leader-1' not in app.components


def test_get_component(app_with_components):
    """Test retrieving a component from an App."""
    app = app_with_components
    component = app.components['test-leader-1']
    assert app.get_component('test-leader-1') == component
    with pytest.raises(KeyError):
        app.get_component('x-invalid-component')


def test_get_components_of_type(app_with_components):
    """Test retrieving list of components of a given type."""
    app = app_with_components
    components = app.get_components_of_type(FollowerComponent)
    follower_components = [
        app.components['test-follower-1'],
        app.components['test-follower-2'],
        app.components['test-leader-1'],
        app.components['test-leader-2'],
    ]
    assert list(components) == follower_components
    components = app.get_components_of_type(LeaderTest)
    leader_components = [
        app.components['test-leader-1'],
        app.components['test-leader-2'],
    ]
    assert list(components) == leader_components


def test_app_is_available(app_with_components):
    """Test checking if an app is available for setup."""
    for component in app_with_components.components.values():
        component.is_available = Mock(return_value=True)

    assert app_with_components.is_available()
    for component in app_with_components.components.values():
        component.is_available.assert_has_calls([call()])

    component = list(app_with_components.components.values())[0]
    component.is_available.return_value = False
    assert not app_with_components.is_available()


def test_app_setup(app_with_components):
    """Test that running setup on an app runs setup on components."""
    for component in app_with_components.components.values():
        component.setup = Mock()

    app_with_components.setup(old_version=2)
    for component in app_with_components.components.values():
        component.setup.assert_has_calls([call(old_version=2)])


def test_app_uninstall(app_with_components):
    """Test that running uninstall on an app runs uninstall on components."""
    for component in app_with_components.components.values():
        component.uninstall = Mock()

    app_with_components.uninstall()
    for component in app_with_components.components.values():
        component.uninstall.assert_has_calls([call()])


@pytest.mark.django_db
def test_setup_state():
    """Test retrieving the current setup state of the app."""
    app = AppSetupTest()
    app.info.version = 3

    app.set_setup_version(3)
    assert app.get_setup_state() == App.SetupState.UP_TO_DATE
    assert not app.needs_setup()

    app.set_setup_version(0)
    assert app.get_setup_state() == App.SetupState.NEEDS_SETUP
    assert app.needs_setup()

    app.set_setup_version(2)
    assert app.get_setup_state() == App.SetupState.NEEDS_UPDATE
    assert not app.needs_setup()


@pytest.mark.django_db
def test_get_set_setup_version():
    """Setting and getting the setup version of the app."""
    app = AppSetupTest()

    from plinth import models
    try:
        models.Module.objects.get(pk=app.app_id).delete()
    except models.Module.DoesNotExist:
        pass
    assert app.get_setup_version() == 0

    app.set_setup_version(5)
    assert app.get_setup_version() == 5


def test_app_enable(app_with_components):
    """Test that enabling an app enables components."""
    app_with_components.disable()
    app_with_components.enable()
    for component in app_with_components.components.values():
        assert component.is_enabled()


def test_app_disable(app_with_components):
    """Test that disabling an app disables components."""
    app_with_components.enable()
    app_with_components.disable()
    for component in app_with_components.components.values():
        assert not component.is_enabled()


def test_app_is_enabled(app_with_components):
    """Test checking for app enabled."""
    app = app_with_components
    app.disable()

    # Disabling the components disables that app
    assert not app.is_enabled()

    # Enabling followers will not enable the app
    app.components['test-follower-1'].enable()
    assert not app.is_enabled()
    app.components['test-follower-2'].enable()
    assert not app.is_enabled()

    # Enabling both leaders will enable the app
    app.components['test-leader-1'].enable()
    assert not app.is_enabled()
    app.components['test-leader-2'].enable()
    assert app.is_enabled()

    # Disabling followers has no effect
    app.components['test-follower-1'].disable()
    assert app.is_enabled()


def test_app_is_enabled_with_no_leader_components():
    """When there are not leader components, app.is_enabled() returns True."""
    app = AppTest()
    assert app.is_enabled()


def test_app_set_enabled(app_with_components):
    """Test that setting enabled effects only followers."""
    app = app_with_components

    app.disable()
    app.set_enabled(True)
    assert app.components['test-follower-1'].is_enabled()
    assert not app.components['test-leader-1'].is_enabled()

    app.enable()
    app.set_enabled(False)
    assert not app.components['test-follower-1'].is_enabled()
    assert app.components['test-leader-1'].is_enabled()


def test_app_diagnose(app_with_components):
    """Test running diagnostics on an app."""
    results = app_with_components.diagnose()
    assert results == [
        DiagnosticCheck('test-result-test-leader-1',
                        'test-result-test-leader-1', Result.PASSED),
        DiagnosticCheck('test-result-test-leader-2',
                        'test-result-test-leader-2', Result.PASSED),
    ]


def test_app_has_diagnostics(app_with_components):
    """Test checking if app has diagnostics implemented."""
    app = app_with_components

    # App with components that has diagnostics
    assert app.has_diagnostics()

    # App with components that don't have diagnostics
    app.remove('test-leader-1')
    app.remove('test-leader-2')
    assert not app.has_diagnostics()

    # App with app-level diagnostics
    with patch.object(
            AppTest, 'diagnose',
            return_value=[DiagnosticCheck('test1', 'test1', Result.PASSED)]):
        assert app.has_diagnostics()


@patch('plinth.setup.run_setup_on_app')
def test_app_repair(_run_setup_on_app, app_with_components):
    """Test running repair on an app."""
    component = app_with_components.get_component('test-follower-1')
    component.repair = Mock(return_value=True)

    check1 = DiagnosticCheck('check1', 'check1', Result.FAILED, {})
    check2 = DiagnosticCheck('check2', 'check2', Result.WARNING, {})
    check3 = DiagnosticCheck('check3', 'check3', Result.FAILED, {},
                             'test-follower-1')
    should_rerun_setup = app_with_components.repair([])
    assert not should_rerun_setup

    should_rerun_setup = app_with_components.repair([check1])
    assert should_rerun_setup

    should_rerun_setup = app_with_components.repair([check2])
    assert should_rerun_setup

    should_rerun_setup = app_with_components.repair([check1, check2])
    assert should_rerun_setup
    component.repair.assert_not_called()

    should_rerun_setup = app_with_components.repair([check3])
    assert should_rerun_setup
    assert component.repair.mock_calls == [call([check3])]

    component.repair = Mock(return_value=False)
    should_rerun_setup = app_with_components.repair([check3])
    assert not should_rerun_setup


def test_app_has_logs(app_with_components):
    """Test checking if an app has logs."""
    app = app_with_components

    # App with components that emit logs
    assert app.has_logs()

    # App with components that don't have diagnostics
    app.remove('test-leader-1')
    app.remove('test-leader-2')
    assert not app.has_diagnostics()


@patch('plinth.privileged.service._assert_service_is_managed_by_plinth')
@patch('plinth.action_utils.service_get_logs')
@patch('plinth.action_utils.service_show')
def test_app_get_logs(service_show, service_get_logs, _, app_with_components,
                      mock_privileged):
    """Test retrieving logs from an app."""
    service_show.side_effect = [{
        'Description': 'Test Desc 1'
    }, {
        'Description': 'Test Desc 2'
    }]
    service_get_logs.side_effect = ['test-logs-1', 'test-logs-2']

    logs = app_with_components.get_logs()
    assert logs == [{
        'unit': 'test-leader-1',
        'description': 'Test Desc 1',
        'logs': 'test-logs-1'
    }, {
        'unit': 'test-leader-2',
        'description': 'Test Desc 2',
        'logs': 'test-logs-2'
    }]
    assert service_show.mock_calls == [
        call('test-leader-1'), call('test-leader-2')
    ]
    assert service_get_logs.mock_calls == [
        call('test-leader-1'), call('test-leader-2')
    ]


def test_component_initialization():
    """Test that component is initialized properly."""
    with pytest.raises(ValueError):
        Component(None)

    component = Component('test-component')
    assert component.component_id == 'test-component'
    assert not component.is_leader


def test_component_app_property():
    """Test component's app property."""
    component = Component('test-component')
    assert component.app_id is None
    with pytest.raises(KeyError):
        assert not component.app

    app = AppTest()
    app.add(component)
    assert component.app == app


def test_component_is_available():
    """Test checking if a component is available for setup."""
    component = Component('test-component')
    assert component.is_available()


def test_component_setup():
    """Test running setup on component."""
    component = Component('test-component')
    assert component.setup(old_version=1) is None


def test_component_uninstall():
    """Test running uninstall on component."""
    component = Component('test-component')
    assert component.uninstall() is None


def test_component_enable():
    """Test running enable on component."""
    component = Component('test-component')
    assert component.enable() is None


def test_component_disable():
    """Test running disable on component."""
    component = Component('test-component')
    assert component.disable() is None


def test_component_diagnose():
    """Test running diagnostics on component."""
    component = Component('test-component')
    assert component.diagnose() == []


def test_component_has_diagnostics():
    """Test checking if component has diagnostics implemented."""
    component = LeaderTest('test-leader-1')
    assert component.has_diagnostics()

    component = FollowerComponent('test-follower-1')
    assert not component.has_diagnostics()


@patch('plinth.setup.run_setup_on_app')
def test_component_repair(_run_setup_on_app):
    """Test running repair on component."""
    component = Component('test-component')
    assert component.repair(['test-check'])


def test_follower_component_initialization():
    """Test that follower component is initialized properly."""
    component = FollowerComponent('test-follower-1')
    assert not component.is_enabled()

    component = FollowerComponent('test-follower-2', False)
    assert not component.is_enabled()

    component = FollowerComponent('test-follower-3', True)
    assert component.is_enabled()


def test_follower_component_set_enabled():
    """Test setting internal enabled state a follower component."""
    component = FollowerComponent('test-follower-1', False)
    component.set_enabled(True)
    assert component.is_enabled()
    component.set_enabled(False)
    assert not component.is_enabled()


def test_follower_component_enable():
    """Test enabling a follower component."""
    component = FollowerComponent('test-follower-1', False)
    component.enable()
    assert component.is_enabled()


def test_follower_component_disable():
    """Test disabling a follower component."""
    component = FollowerComponent('test-follower-1', True)
    component.disable()
    assert not component.is_enabled()


def test_leader_component_initialization():
    """Test that leader component is initialized properly."""
    component = LeaderComponent('test-leader-1')
    assert component.is_leader


def test_leader_component_is_enabled():
    """Test getting enabled state is not implemented in leader component."""
    component = LeaderComponent('test-leader-1')
    with pytest.raises(NotImplementedError):
        assert component.is_enabled()


def test_info_initialization_without_args():
    """Test initializing the Info component without arguments."""
    info = Info('test-app', 3)
    assert info.component_id == 'test-app-info'
    assert info.app_id == 'test-app'
    assert info.version == 3
    assert not info.is_essential
    assert info.depends == []
    assert info.name is None
    assert info.icon is None
    assert info.icon_filename is None
    assert info.description is None
    assert info.manual_page is None
    assert info.clients is None
    assert info.tags == []


def test_info_initialization_with_args():
    """Test initializing the Info component with arguments."""
    clients = [{
        'name': 'test',
        'platforms': [{
            'type': 'web',
            'url': 'test-url'
        }]
    }]
    info = Info('test-app', 3, is_essential=True, depends=['test-app-2'],
                name='Test App', icon='fa-test', icon_filename='test-icon',
                description='Test description', manual_page='Test',
                clients=clients, tags=['tag1', 'tag2'])
    assert info.is_essential
    assert info.depends == ['test-app-2']
    assert info.name == 'Test App'
    assert info.icon == 'fa-test'
    assert info.icon_filename == 'test-icon'
    assert info.description == 'Test description'
    assert info.manual_page == 'Test'
    assert info.clients == clients
    assert info.tags == ['tag1', 'tag2']


def test_info_clients_validation():
    """Test clients parameter validation during initialization."""
    with pytest.raises(AssertionError):
        Info('test-app', 3, clients='invalid')

    clients = [{
        'name': 'test',
        'platforms': [{
            'type': 'web',
            'url': 'test-url'
        }]
    }]
    Info('test-app', 3, clients=clients)


def test_enable_state_key(app_with_components):
    """Test getting the storage key for enable state component."""
    component = EnableState('enable-state-1')
    with pytest.raises(KeyError):
        assert component.key

    app_with_components.add(component)
    assert component.key == app_with_components.app_id + '_enable'


@pytest.mark.django_db
def test_enable_state_enable_disable(app_with_components):
    """Test enabling/disabling enable state component."""
    component = EnableState('enable-state-1')
    app_with_components.add(component)
    assert not component.is_enabled()
    component.enable()
    assert component.is_enabled()
    component.disable()
    assert not component.is_enabled()


class ModuleTest1:
    """A test module with an app."""

    class App1(App):
        """A non-essential app that depends on another."""
        app_id = 'app1'

        def __init__(self):
            super().__init__()
            self.add(Info('app1', version=1, depends=['app3']))


class ModuleTest2:
    """A test module with multiple apps."""

    class App2(App):
        """An essential app."""
        app_id = 'app2'

        def __init__(self):
            super().__init__()
            self.add(Info('app2', version=1, is_essential=True))

    class App3(App):
        """An non-essential app that is depended on by another."""
        app_id = 'app3'

        def __init__(self):
            super().__init__()
            self.add(Info('app3', version=1))


@patch('plinth.module_loader.loaded_modules')
def test_apps_init(loaded_modules):
    """Test that initializing all apps works."""
    loaded_modules.items.return_value = [('test1', ModuleTest1()),
                                         ('test2', ModuleTest2())]

    apps_init()
    assert list(App._all_apps.keys()) == ['app2', 'app3', 'app1']


class ModuleCircularTest:
    """A test module with apps depending on each other."""

    class App1(App):
        """An app depending on app2."""
        app_id = 'app1'

        def __init__(self):
            super().__init__()
            self.add(Info('app1', version=1, depends=['app2']))

    class App2(App):
        """An app depending on app1."""
        app_id = 'app2'

        def __init__(self):
            super().__init__()
            self.add(Info('app2', version=1, depends=['app1']))

    class App3(App):
        """An app without dependencies."""
        app_id = 'app3'

        def __init__(self):
            super().__init__()
            self.add(Info('app3', version=1))


@patch('plinth.module_loader.loaded_modules')
def test_apps_init_circular_depends(loaded_modules):
    """Test initializing apps with circular dependencies."""
    loaded_modules.items.return_value = [('test', ModuleCircularTest())]

    apps_init()
    assert list(App._all_apps.keys()) == ['app3']



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Minetest server."""

import pathlib
import shutil

import augeas

from plinth import action_utils
from plinth.actions import privileged

old_config_file = pathlib.Path('/etc/minetest/minetest.conf')
config_file = pathlib.Path('/etc/luanti/default.conf')
AUG_PATH = '/files' + str(config_file) + '/.anon'


@privileged
def setup() -> None:
    """Migrate old configuration file."""
    if old_config_file.exists():
        old_config_file.rename(config_file)
        action_utils.service_daemon_reload()
        action_utils.service_try_restart('luanti-server')


@privileged
def configure(max_players: int | None = None, enable_pvp: bool | None = None,
              creative_mode: bool | None = None,
              enable_damage: bool | None = None):
    """Update configuration file and restart daemon if necessary."""
    config_file.parent.mkdir(exist_ok=True)
    aug = load_augeas()
    if max_players is not None:
        aug.set(AUG_PATH + '/max_users', str(max_players))

    if enable_pvp is not None:
        aug.set(AUG_PATH + '/enable_pvp', str(enable_pvp).lower())

    if creative_mode is not None:
        aug.set(AUG_PATH + '/creative_mode', str(creative_mode).lower())

    if enable_damage is not None:
        aug.set(AUG_PATH + '/enable_damage', str(enable_damage).lower())

    aug.save()
    action_utils.service_try_restart('luanti-server')


def load_augeas():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.set('/augeas/load/Php/lens', 'Php.lns')
    aug.set('/augeas/load/Php/incl[last() + 1]', str(config_file))
    aug.load()
    return aug


@privileged
def uninstall() -> None:
    """Remove the data directory that luanti-server package fails to remove.

    See: https://bugs.debian.org/1122677
    """
    shutil.rmtree('/var/lib/private/luanti/default/')



# SPDX-License-Identifier: AGPL-3.0-or-later

#
# Generated by Django 4.2.28 on 2026-03-19 23:52
#
"""
Django migration for adding the user's passkey model.
"""

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models


class Migration(migrations.Migration):

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('plinth', '0005_storednotification'),
    ]

    operations = [
        migrations.CreateModel(
            name='UserPasskey',
            fields=[
                ('id',
                 models.AutoField(auto_created=True, primary_key=True,
                                  serialize=False, verbose_name='ID')),
                ('name', models.CharField(max_length=256, null=True,
                                          blank=True)),
                ('domain', models.CharField(max_length=256)),
                ('created_time', models.DateTimeField(auto_now_add=True)),
                ('last_used_time', models.DateTimeField(auto_now=True)),
                ('signature_counter', models.IntegerField(default=0)),
                ('registration_flags', models.IntegerField()),
                ('extensions', models.JSONField(null=True, blank=True)),
                ('aaguid', models.UUIDField(null=True, blank=True)),
                ('credential_id', models.BinaryField(unique=True)),
                ('public_key', models.BinaryField()),
                ('user',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                   related_name='passkeys',
                                   to=settings.AUTH_USER_MODEL)),
            ],
        ),
    ]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
App component for other apps to use firewall functionality.
"""

import logging
import re
from typing import ClassVar, TypeAlias

from django.utils.translation import gettext_noop

from plinth import app
from plinth.diagnostic_check import (DiagnosticCheck,
                                     DiagnosticCheckParameters, Result)
from plinth.modules import firewall

logger = logging.getLogger(__name__)

_list_type: TypeAlias = list


class Firewall(app.FollowerComponent):
    """Component to open/close firewall ports for an app."""

    _all_firewall_components: ClassVar[dict[str, 'Firewall']] = {}

    def __init__(self, component_id, name=None, ports=None, is_external=False):
        """Initialize the firewall component."""
        super().__init__(component_id)

        if not ports:
            ports = []

        self.name = name
        self.ports = ports
        self.is_external = is_external

        self._all_firewall_components[component_id] = self

    @property
    def ports_details(self):
        """Retrieve details of ports associated with this component.."""
        ports_details = []
        for port in self.ports:
            ports_details.append({
                'name': port,
                'details': firewall.get_port_details(port),
            })

        return ports_details

    @classmethod
    def list(cls):
        """Return a list of all firewall ports."""
        return cls._all_firewall_components.values()

    def enable(self):
        """Open firewall ports when the component is enabled."""
        super().enable()
        firewall.try_with_reload(self._enable)

    def _enable(self):
        """Open firewall ports."""
        internal_enabled_ports = firewall.get_enabled_services(zone='internal')
        external_enabled_ports = firewall.get_enabled_services(zone='external')

        logger.info('Firewall ports opened - %s, %s', self.name, self.ports)
        for port in self.ports:
            if port not in internal_enabled_ports:
                firewall.add_service(port, zone='internal')

            if (self.is_external and port not in external_enabled_ports):
                firewall.add_service(port, zone='external')

    def disable(self):
        """Close firewall ports when the component is disabled."""
        super().disable()
        firewall.try_with_reload(self._disable)

    def _disable(self):
        """Close firewall ports."""
        internal_enabled_ports = firewall.get_enabled_services(zone='internal')
        external_enabled_ports = firewall.get_enabled_services(zone='external')

        logger.info('Firewall ports closed - %s, %s', self.name, self.ports)
        for port in self.ports:
            if port in internal_enabled_ports:
                enabled_components_on_port = [
                    component.is_enabled()
                    for component in self._all_firewall_components.values()
                    if port in component.ports
                    and self.component_id != component.component_id
                ]
                if not any(enabled_components_on_port):
                    firewall.remove_service(port, zone='internal')

            if port in external_enabled_ports:
                enabled_components_on_port = [
                    component.is_enabled()
                    for component in self._all_firewall_components.values()
                    if port in component.ports and self.component_id !=
                    component.component_id and component.is_external
                ]
                if not any(enabled_components_on_port):
                    firewall.remove_service(port, zone='external')

    @staticmethod
    def get_internal_interfaces():
        """Returns a list of interfaces in a firewall zone.

        Filter out tun interfaces as they are always assumed to be internal
        interfaces.

        """
        return [
            interface for interface in firewall.get_interfaces('internal')
            if not re.fullmatch(r'tun\d+', interface)
        ]

    def diagnose(self) -> _list_type[DiagnosticCheck]:
        """Check if the firewall ports are open and only as expected.

        See :py:meth:`plinth.app.Component.diagnose`.

        """
        results = []
        internal_ports = firewall.get_enabled_services(zone='internal')
        external_ports = firewall.get_enabled_services(zone='external')
        for port_detail in self.ports_details:
            port = str(port_detail['name'])
            details = ', '.join(
                (f'{port_number}/{protocol}'
                 for port_number, protocol in port_detail['details']))

            # Internal zone
            check_id = f'firewall-port-internal-{port}'
            result = Result.PASSED if port in internal_ports else Result.FAILED
            description = gettext_noop(
                'Port {name} ({details}) available for internal networks')
            parameters: DiagnosticCheckParameters = {
                'name': port,
                'details': details
            }
            results.append(
                DiagnosticCheck(check_id, description, result, parameters,
                                self.component_id))

            # External zone
            if self.is_external:
                check_id = f'firewall-port-external-available-{port}'
                result = Result.PASSED \
                    if port in external_ports else Result.FAILED
                description = gettext_noop(
                    'Port {name} ({details}) available for external networks')
            else:
                check_id = f'firewall-port-external-unavailable-{port}'
                result = Result.PASSED \
                    if port not in external_ports else Result.FAILED
                description = gettext_noop(
                    'Port {name} ({details}) unavailable for external networks'
                )

            parameters = {'name': port, 'details': details}
            results.append(
                DiagnosticCheck(check_id, description, result, parameters,
                                self.component_id))

        return results


class FirewallLocalProtection(app.FollowerComponent):
    """Component to protect local services from access by local users.

    Local service protection means that only administrators and Apache web
    server should be able to access certain services and not other users who
    have logged into the system. This is needed because some of the services
    are protected with authentication and authorization provided by Apache web
    server. If services are contacted directly then auth can be bypassed by all
    local users.

    `component_id` should be a unique ID across all components of an app and
    across all components.

    `tcp_ports` is list of all local TCP ports on which daemons of this app are
    listening. Administrators and Apache web server will be allowed to connect
    and all other connections to these ports will be rejected.
    """

    def __init__(self, component_id: str, tcp_ports: list[str]):
        """Initialize the firewall component."""
        super().__init__(component_id)

        self.tcp_ports = tcp_ports

    def enable(self):
        """Block traffic to local service from local users."""
        super().enable()
        for port in self.tcp_ports:
            firewall.add_passthrough('ipv6', '-A', 'INPUT', '-p', 'tcp',
                                     '--dport', port, '-j', 'REJECT')
            firewall.add_passthrough('ipv4', '-A', 'INPUT', '-p', 'tcp',
                                     '--dport', port, '-j', 'REJECT')

    def disable(self):
        """Unblock traffic to local service from local users."""
        super().disable()
        for port in self.tcp_ports:
            firewall.remove_passthrough('ipv6', '-A', 'INPUT', '-p', 'tcp',
                                        '--dport', port, '-j', 'REJECT')
            firewall.remove_passthrough('ipv4', '-A', 'INPUT', '-p', 'tcp',
                                        '--dport', port, '-j', 'REJECT')

    def setup(self, old_version):
        """Protect services of an app that newly introduced the feature."""
        if not old_version:
            # Fresh installation of an app. app.enable() will run at the end.
            return

        if self.app.is_enabled():
            # Don't enable if the app is being updated but is disabled.
            self.enable()


def get_port_forwarding_info(app_):
    """Return a list of ports to be forwarded for this app to work."""
    from plinth.modules import networks
    info = {
        'network_topology_type': networks.get_network_topology_type(),
        'router_configuration_type': networks.get_router_configuration_type(),
        'ports': []
    }
    for component in app_.components.values():
        if not isinstance(component, Firewall):
            continue

        if not component.is_external:
            continue

        for port in component.ports_details:
            if port['name'] in ['http', 'https']:
                continue

            for detail in port['details']:
                info['ports'].append({
                    'name': port['name'],
                    'protocol': detail[1].upper(),
                    'ports': detail[0]
                })

    return info



# SPDX-License-Identifier: AGPL-3.0-or-later

from django import forms
from django.core.exceptions import ValidationError
from django.utils.translation import gettext_lazy as _


class UserCredentialsForm(forms.Form):
    """Form to create admin user or change a user's password."""

    username = forms.CharField(label=_('Username'),
                               help_text=_('Enter a username for the user.'))
    password = forms.CharField(
        label=_('Password'), widget=forms.PasswordInput, min_length=6,
        strip=False,
        help_text=_('Enter a strong password with a minimum of 6 characters.'))
    password_confirmation = forms.CharField(
        label=_('Password confirmation'), widget=forms.PasswordInput,
        min_length=6, strip=False,
        help_text=_('Enter the same password for confirmation.'))

    def clean(self):
        """Raise error if passwords don't match."""
        cleaned_data = super().clean()
        password = self.cleaned_data.get('password')
        password_confirmation = self.cleaned_data.get('password_confirmation')

        if password and password_confirmation and (password
                                                   != password_confirmation):
            self.add_error('password_confirmation',
                           ValidationError(_('Passwords do not match.')))

        return cleaned_data



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure MediaWiki."""

import os
import pathlib
import shutil
import subprocess
import tempfile

from plinth import action_utils
from plinth.actions import privileged, secret_str
from plinth.utils import generate_password

MAINTENANCE_SCRIPTS_DIR = "/usr/share/mediawiki/maintenance"
CONF_FILE = '/etc/mediawiki/FreedomBoxSettings.php'
LOCAL_SETTINGS_CONF = '/etc/mediawiki/LocalSettings.php'


def get_php_command():
    """Return the PHP command that should be used on CLI.

    This is workaround for /usr/bin/php pointing to a different version than
    what php-defaults (and php-mbstring, php-xml) point to. See:
    https://bugs.debian.org/cgi-bin/bugreport.cgi?bug=959742

    """
    version = ''

    try:
        process = action_utils.run(['dpkg', '-s', 'php'], check=True)
        for line in process.stdout.decode().splitlines():
            if line.startswith('Version:'):
                version = line.split(':')[-1].split('+')[0].strip()
    except subprocess.CalledProcessError:
        pass

    return f'php{version}'


@privileged
def setup():
    """Run the installer script to create database and configuration file."""
    data_dir = '/var/lib/mediawiki-db/'
    if not os.path.exists(data_dir):
        os.mkdir(data_dir)

    if not os.path.exists(os.path.join(data_dir, 'my_wiki.sqlite')) or \
       not os.path.exists(LOCAL_SETTINGS_CONF):
        install_script = os.path.join(MAINTENANCE_SCRIPTS_DIR, 'install.php')
        password = generate_password()
        with tempfile.NamedTemporaryFile() as password_file_handle:
            password_file_handle.write(password.encode())
            password_file_handle.flush()
            action_utils.run([
                get_php_command(), install_script, '--confpath=/etc/mediawiki',
                '--dbtype=sqlite', '--dbpath=' + data_dir,
                '--scriptpath=/mediawiki', '--passfile',
                password_file_handle.name, 'Wiki', 'admin'
            ], check=True)
    action_utils.run(['chmod', '-R', 'o-rwx', data_dir], check=True)
    action_utils.run(['chown', '-R', 'www-data:www-data', data_dir],
                     check=True)

    conf_file = pathlib.Path(CONF_FILE)
    if not conf_file.exists():
        conf_file.write_text('<?php\n')

    _include_custom_config()


def _include_custom_config():
    """Include FreedomBox specific configuration in LocalSettings.php."""
    with open(LOCAL_SETTINGS_CONF, 'r', encoding='utf-8') as conf_file:
        lines = conf_file.readlines()

    static_settings_index = None
    settings_index = None
    for line_number, line in enumerate(lines):
        if 'FreedomBoxSettings.php' in line:
            settings_index = line_number

        if 'FreedomBoxStaticSettings.php' in line:
            static_settings_index = line_number

    if settings_index is None:
        settings_index = len(lines)
        lines.append('include dirname(__FILE__)."/FreedomBoxSettings.php";\n')

    if static_settings_index is None:
        lines.insert(
            settings_index,
            'include dirname(__FILE__)."/FreedomBoxStaticSettings.php";\n')

    with open(LOCAL_SETTINGS_CONF, 'w', encoding='utf-8') as conf_file:
        conf_file.writelines(lines)


@privileged
def change_password(username: str, password: secret_str):
    """Change the password for a given user."""
    change_password_script = os.path.join(MAINTENANCE_SCRIPTS_DIR,
                                          'changePassword.php')

    action_utils.run([
        get_php_command(), change_password_script, '--user', username,
        '--password', password
    ], check=True)


@privileged
def update():
    """Run update.php maintenance script when version upgrades happen."""
    update_script = os.path.join(MAINTENANCE_SCRIPTS_DIR, 'update.php')
    action_utils.run([get_php_command(), update_script, '--quick'], check=True)


def _update_setting(setting_name, setting_line):
    """Update the value of one setting in the config file."""
    with open(CONF_FILE, 'r', encoding='utf-8') as conf_file:
        lines = conf_file.readlines()

        inserted = False
        for i, line in enumerate(lines):
            if line.strip().startswith(setting_name):
                lines[i] = setting_line
                inserted = True
                break

        if not inserted:
            lines.append(setting_line)

    with open(CONF_FILE, 'w', encoding='utf-8') as conf_file:
        conf_file.writelines(lines)


@privileged
def set_public_registrations(should_enable: bool):
    """Enable or Disable public registrations for MediaWiki."""
    setting = "$wgGroupPermissions['*']['createaccount']"
    conf_value = 'true' if should_enable else 'false'
    _update_setting(setting, f'{setting} = {conf_value};\n')


@privileged
def set_private_mode(should_enable: bool):
    """Enable or Disable Private mode for wiki."""
    setting = "$wgGroupPermissions['*']['read']"
    conf_value = 'false' if should_enable else 'true'
    _update_setting(setting, f'{setting} = {conf_value};\n')


@privileged
def set_default_skin(skin: str):
    """Set a default skin."""
    _update_setting('$wgDefaultSkin', f'$wgDefaultSkin = "{skin}";\n')


@privileged
def set_server_url(server_url: str):
    """Set the value of $wgServer for this MediaWiki server."""
    # This is a required setting from MediaWiki 1.34
    _update_setting('$wgServer', f'$wgServer = "{server_url}";\n')


@privileged
def set_site_name(site_name: str):
    """Set the value of $wgSitename for this MediaWiki server."""
    _update_setting('$wgSitename', f'$wgSitename = "{site_name}";\n')


@privileged
def set_default_language(language: str):
    """Set the value of $wgLanguageCode for this MediaWiki server."""
    _update_setting('$wgLanguageCode ', f'$wgLanguageCode = "{language}";\n')
    # In MediaWiki 1.8 or older, if you change this after installation, you
    # should run the maintenance/rebuildmessages.php script to rebuild the
    # user interface messages (MediaWiki namespace). Otherwise, you will not
    # see the interface in the new language, or a mix of the old and new
    # languages.
    rebuild_messages_script = os.path.join(MAINTENANCE_SCRIPTS_DIR,
                                           'rebuildmessages.php')
    action_utils.run([get_php_command(), rebuild_messages_script], check=True)


@privileged
def uninstall():
    """Remove Mediawiki's database and local config file."""
    shutil.rmtree('/var/lib/mediawiki-db', ignore_errors=True)
    # /etc/mediawiki is removed on purge of mediawiki package



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Home Assistant."""

import functools
import pathlib
import time
import traceback
from dataclasses import dataclass

import yaml

from plinth.actions import privileged

_settings_file = pathlib.Path(
    '/var/lib/home-assistant-freedombox/config/configuration.yaml')


@dataclass
class YAMLUnknownTag:
    """Object used to represent an unknown tag in YAML."""
    tag: str
    value: str


class YAMLLoader(yaml.SafeLoader):
    """Custom YAML loader to handle !include etc. tags."""
    pass


def yaml_unknown_constructor(loader, node, tag):
    """Create an object when a unknown tag is encountered."""
    value = loader.construct_scalar(node)
    return YAMLUnknownTag(tag, value)


class YAMLDumper(yaml.Dumper):
    """Custom YAML dumper to handle !include etc. tags."""
    pass


def yaml_unknown_representor(dumper, data):
    """Dump original tag from an object representing an unknown tag."""
    return dumper.represent_scalar(data.tag, data.value)


def yaml_add_handlers():
    """Add special handlers to YAML loader and dumper."""
    tags = [
        '!include', '!env_var', '!secret', '!include_dir_list',
        '!include_dir_merge_list', '!include_dir_named',
        '!include_dir_merge_named', '!input'
    ]
    for tag in tags:
        YAMLLoader.add_constructor(
            tag, functools.partial(yaml_unknown_constructor, tag=tag))

    YAMLDumper.add_representer(YAMLUnknownTag, yaml_unknown_representor)


yaml_add_handlers()


@privileged
def setup() -> None:
    """Setup basic Home Assistant configuration."""
    pathlib.Path('/var/lib/home-assistant-freedombox/').chmod(0o700)

    try:
        _wait_for_configuration_file()

        settings = _read_settings()
        if 'http' not in settings:
            settings['http'] = {}

        settings['http']['server_host'] = '127.0.0.1'
        settings['http']['use_x_forwarded_for'] = True
        settings['http']['trusted_proxies'] = '127.0.0.1'
        _write_settings(settings)
    except Exception as exception:
        raise Exception(
            traceback.format_tb(exception.__traceback__) +
            [_settings_file.read_text()])


def _wait_for_configuration_file() -> None:
    """Wait until the Home Assistant daemon creates a configuration file."""
    start_time = time.time()
    while time.time() < start_time + 300:
        if _settings_file.exists():
            break

        time.sleep(1)


def _read_settings() -> dict:
    """Load settings as dictionary from YAML config file."""
    with _settings_file.open('rb') as settings_file:
        return yaml.load(settings_file, Loader=YAMLLoader)


def _write_settings(settings: dict):
    """Write settings from dictionary to YAML config file."""
    with _settings_file.open('w', encoding='utf-8') as settings_file:
        yaml.dump(settings, settings_file, Dumper=YAMLDumper)



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

from plinth.clients import store_url

_element_android_package_id = 'im.vector.app'
_element_desktop_download_url = 'https://element.io/get-started'
_fluffychat_android_package_id = 'chat.fluffy.fluffychat'
_fluffychat_desktop_download_url = 'https://fluffychat.im/'

clients = [{
    'name':
        _('Element'),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', _element_android_package_id)
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', _element_android_package_id)
    }, {
        'type': 'store',
        'os': 'ios',
        'store_name': 'app-store',
        'url': 'https://apps.apple.com/app/vector/id1083446067'
    }, {
        'type': 'web',
        'url': 'https://app.element.io/'
    }, {
        'type': 'download',
        'os': 'gnu-linux',
        'url': _element_desktop_download_url,
    }, {
        'type': 'download',
        'os': 'macos',
        'url': _element_desktop_download_url,
    }, {
        'type': 'download',
        'os': 'windows',
        'url': _element_desktop_download_url,
    }]
}, {
    'name':
        _('FluffyChat'),
    'platforms': [{
        'type': 'store',
        'os': 'android',
        'store_name': 'google-play',
        'url': store_url('google-play', _fluffychat_android_package_id)
    }, {
        'type': 'store',
        'os': 'android',
        'store_name': 'f-droid',
        'url': store_url('f-droid', _fluffychat_android_package_id)
    }, {
        'type': 'store',
        'os': 'ios',
        'store_name': 'app-store',
        'url': 'https://apps.apple.com/app/fluffychat/id1551469600'
    }, {
        'type': 'web',
        'url': 'https://fluffychat.im/web'
    }, {
        'type': 'download',
        'os': 'gnu-linux',
        'url': _fluffychat_desktop_download_url,
    }]
}]

backup = {
    'config': {
        'directories': ['/etc/matrix-synapse/conf.d/'],
        'files': [
            '/etc/matrix-synapse/homeserver.yaml',
            '/etc/matrix-synapse/log.yaml'
        ]
    },
    'data': {
        'directories': [
            '/var/lib/matrix-synapse/media/',
            '/var/lib/matrix-synapse/uploads/'
        ],
        'files': ['/var/lib/matrix-synapse/homeserver.db']
    },
    'secrets': {
        'files': [
            '/etc/matrix-synapse/homeserver.signing.key',
            '/etc/matrix-synapse/homeserver.tls.crt',
            '/etc/matrix-synapse/homeserver.tls.dh',
            '/etc/matrix-synapse/homeserver.tls.key'
        ]
    },
    'services': ['matrix-synapse']
}

tags = [
    _('Chat room'),
    _('Encrypted messaging'),
    _('Audio chat'),
    _('Video chat'),
    _('Matrix server'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later

from django.utils.translation import gettext_lazy as _

clients = [{
    'name': _('bepasty'),
    'platforms': [{
        'type': 'web',
        'url': '/bepasty'
    }]
}]

backup = {
    'config': {
        'files': ['/etc/bepasty-freedombox.conf']
    },
    'data': {
        'directories': ['/var/lib/bepasty', '/var/lib/private/bepasty']
    },
    'services': ['uwsgi-app@bepasty-freedombox.socket'],
}

tags = [_('File sharing'), _('Pastebin')]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for Zoph server."""

import configparser
import pathlib
import re
import subprocess

from plinth import action_utils
from plinth.actions import privileged
from plinth.db import mariadb

APACHE_CONF = '/etc/apache2/conf-available/zoph.conf'
DB_CONF = pathlib.Path('/etc/zoph.ini')
DB_BACKUP_FILE = '/var/lib/plinth/backups-data/zoph-database.sql'


@privileged
def pre_install():
    """Preseed debconf values before packages are installed."""
    action_utils.debconf_set_selections([
        'zoph zoph/dbconfig-install boolean true',
        'zoph zoph/dbconfig-upgrade boolean true',
        'zoph zoph/dbconfig-remove boolean true',
        'zoph zoph/dbconfig-reinstall boolean true'
        'zoph zoph/mysql/admin-user string root',
        'zoph zoph/rm_images select yes',
    ])


@privileged
def get_configuration() -> dict[str, str]:
    """Return the current configuration."""
    configuration = {}
    try:
        process = action_utils.run(['zoph', '--dump-config'], check=True)
    except subprocess.CalledProcessError as exception:
        if exception.returncode != 96:
            raise

        _zoph_setup_cli_user()
        process = action_utils.run(['zoph', '--dump-config'], check=True)

    for line in process.stdout.decode().splitlines():
        name, value = line.partition(':')[::2]
        configuration[name.strip()] = value[1:]

    return configuration


def _zoph_setup_cli_user() -> None:
    """Ensure that Zoph cli user is not set to 'autodetect'.

    When set to 'autodetect', all command line commands will fail unless a user
    name 'root' exists in zoph database and is an admin user.
    """
    query = '''
UPDATE
  zoph_conf
SET
  value=(
    SELECT user_id
    FROM zoph_users
    WHERE user_class="0"
    ORDER BY user_id
    LIMIT 1)
WHERE
  conf_id='interface.user.cli';'''
    database_name = _get_db_config()['db_name']
    mariadb.run_query(database_name, query)


def _zoph_configure(key, value):
    """Set a configure value in Zoph."""
    try:
        action_utils.run(['zoph', '--config', key, value], check=True)
    except subprocess.CalledProcessError as exception:
        if exception.returncode != 96:
            raise

        _zoph_setup_cli_user()
        action_utils.run(['zoph', '--config', key, value], check=True)


@privileged
def setup():
    """Setup Zoph configuration.

    May be called when app is disabled.
    """
    with action_utils.service_ensure_running('mysql'):
        _zoph_configure('import.enable', 'true')
        _zoph_configure('import.upload', 'true')
        _zoph_configure('import.rotate', 'true')
        _zoph_configure('path.unzip', 'unzip')
        _zoph_configure('path.untar', 'tar xvf')
        _zoph_configure('path.ungz', 'gunzip')

        # Maps using OpenStreetMap is enabled by default.
        _zoph_configure('maps.provider', 'osm')


def _get_db_config():
    """Return the name of the database configured by dbconfig."""
    config = configparser.ConfigParser()
    with DB_CONF.open('r', encoding='utf-8') as file_handle:
        config.read_file(file_handle)

    return {
        'db_host': config['zoph']['db_host'].strip('"'),
        'db_name': config['zoph']['db_name'].strip('"'),
        'db_user': config['zoph']['db_user'].strip('"'),
        'db_pass': config['zoph']['db_pass'].strip('"'),
    }


@privileged
def set_configuration(enable_osm: bool | None = None,
                      admin_user: str | None = None):
    """Setup Zoph Apache configuration."""
    _zoph_configure('interface.user.remote', 'true')

    # Note that using OpenSteetmap as a mapping provider is a very nice
    # feature, but some people may regard its use as a privacy issue
    if enable_osm is not None:
        value = 'osm' if enable_osm else ''
        _zoph_configure('maps.provider', value)

    if admin_user:
        # Edit the database to rename the admin user to FreedomBox admin user.
        if not re.match(r'^[\w.@][\w.@-]+\Z', admin_user, flags=re.ASCII):
            # Check to avoid SQL injection
            raise ValueError('Invalid username')

        query = f"UPDATE zoph_users SET user_name='{admin_user}' \
                 WHERE user_name='admin';"

        action_utils.run(['mysql', _get_db_config()['db_name']],
                         input=query.encode(), check=True)


@privileged
def is_configured() -> bool | None:
    """Return whether zoph app is configured."""
    try:
        process = action_utils.run(
            ['zoph', '--get-config', 'interface.user.remote'], check=True)
        return process.stdout.decode().strip() == 'true'
    except (FileNotFoundError, subprocess.CalledProcessError):
        return None


@privileged
def dump_database():
    """Dump database to file.

    May be called when app is disabled.
    """
    with action_utils.service_ensure_running('mysql'):
        db_name = _get_db_config()['db_name']
        with open(DB_BACKUP_FILE, 'w', encoding='utf-8') as db_backup_file:
            action_utils.run(['mysqldump', db_name], stdout=db_backup_file,
                             check=True)


@privileged
def restore_database():
    """Restore database from file.

    May be called when app is disabled.
    """
    with action_utils.service_ensure_running('mysql'):
        db_name = _get_db_config()['db_name']
        db_user = _get_db_config()['db_user']
        db_host = _get_db_config()['db_host']
        db_pass = _get_db_config()['db_pass']
        action_utils.run(['mysqladmin', '--force', 'drop', db_name],
                         check=False)
        action_utils.run(['mysqladmin', 'create', db_name], check=True)
        with open(DB_BACKUP_FILE, 'r', encoding='utf-8') as db_restore_file:
            action_utils.run(['mysql', db_name], stdin=db_restore_file,
                             check=True)

        # Set the password for user from restored configuration
        query = f'ALTER USER {db_user}@{db_host} IDENTIFIED BY "{db_pass}";'
        action_utils.run(['mysql'], input=query.encode(), check=True)


@privileged
def uninstall():
    """Drop database, database user and database configuration.

    May be called when app is disabled.
    """
    with action_utils.service_ensure_running('mysql'):
        try:
            config = _get_db_config()
            action_utils.run(
                ['mysqladmin', '--force', 'drop', config['db_name']],
                check=False)

            query = f'DROP USER IF EXISTS {config["db_user"]}@localhost;'
            action_utils.run(['mysql'], input=query.encode(), check=False)
        except FileNotFoundError:  # Database configuration not found
            pass

    DB_CONF.unlink(missing_ok=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for datetime app.
"""

import pytest
from plinth.tests import functional

pytestmark = [pytest.mark.system, pytest.mark.essential, pytest.mark.datetime]


@pytest.fixture(scope='module', autouse=True)
def fixture_background(session_browser):
    """Login."""
    functional.login(session_browser)


def test_enable_disable(session_browser):
    """Test enabling the app."""
    if functional.app_can_be_disabled(session_browser, 'datetime'):
        functional.app_disable(session_browser, 'datetime')

        functional.app_enable(session_browser, 'datetime')
        assert functional.service_is_running(session_browser, 'datetime')

        functional.app_disable(session_browser, 'datetime')
        assert functional.service_is_not_running(session_browser, 'datetime')

    else:
        pytest.skip('datetime cannot be disabled.')


def test_set_timezone(session_browser):
    """Test setting the timezone."""
    _time_zone_set(session_browser, 'Africa/Abidjan')
    assert _time_zone_get(session_browser) == 'Africa/Abidjan'


@pytest.mark.backups
def test_backup_and_restore(session_browser):
    """Test backup and restore of datetime settings."""
    _time_zone_set(session_browser, 'Africa/Accra')
    functional.backup_create(session_browser, 'datetime', 'test_datetime')

    _time_zone_set(session_browser, 'Africa/Cairo')
    functional.backup_restore(session_browser, 'datetime', 'test_datetime')

    assert _time_zone_get(session_browser) == 'Africa/Accra'


def _time_zone_set(browser, time_zone):
    """Set the system time zone."""
    functional.nav_to_module(browser, 'datetime')
    browser.select('time_zone', time_zone)
    functional.submit(browser, form_class='form-configuration')


def _time_zone_get(browser):
    """Set the system time zone."""
    functional.nav_to_module(browser, 'datetime')
    return browser.find_by_name('time_zone').first.value



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for users views.
"""

from unittest.mock import Mock, patch

import pytest
from django import urls
from django.contrib.auth.models import User
from django.contrib.messages.storage.fallback import FallbackStorage
from django.core.exceptions import PermissionDenied

from plinth import module_loader
from plinth.modules.users import views

from ..components import UsersAndGroups

# For all tests, plinth.urls instead of urls configured for testing, and
# django database
pytestmark = [pytest.mark.urls('plinth.urls'), pytest.mark.django_db]


@pytest.fixture(autouse=True, scope='module')
def fixture_users_urls():
    """Make sure users app's URLs are part of plinth.urls."""
    with patch('plinth.module_loader._modules_to_load', new=[]) as modules, \
            patch('plinth.urls.urlpatterns', new=[]):
        modules.append('plinth.modules.users')
        module_loader.include_urls()
        yield


@pytest.fixture(autouse=True)
def module_patch():
    """Patch users module."""
    pwd_users = [Mock(pw_name='root'), Mock(pw_name='plinth')]

    UsersAndGroups._all_components = set()
    UsersAndGroups('test-users-and-groups',
                   groups={'admin': 'The admin group'})
    UsersAndGroups('users-and-groups-minetest',
                   reserved_usernames=['debian-minetest'])

    privileged = 'plinth.modules.users.privileged'
    with patch('pwd.getpwall', return_value=pwd_users), \
         patch(f'{privileged}.create_user'), \
         patch(f'{privileged}.add_user_to_group'), \
         patch(f'{privileged}.set_user_password'), \
         patch(f'{privileged}.set_user_status'), \
         patch(f'{privileged}.rename_user'), \
         patch(f'{privileged}.get_group_users') as get_group_users, \
         patch('plinth.modules.ssh.privileged.set_keys'), \
         patch('plinth.modules.ssh.privileged.get_keys') as get_keys, \
         patch(f'{privileged}.get_user_groups') as get_user_groups, \
         patch(f'{privileged}.remove_user'):
        get_group_users.return_value = ['admin']
        get_keys.return_value = []
        get_user_groups.return_value = []
        yield


def make_request(request, view, as_admin=True, **kwargs):
    """Make request with a message storage and users created."""
    setattr(request, 'session', 'session')
    messages = FallbackStorage(request)
    setattr(request, '_messages', messages)

    user = User.objects.create(username='tester')
    admin_user = User.objects.create(username='admin')
    admin_user.set_password('adminpassword')

    request.user = admin_user if as_admin else user

    with patch('plinth.modules.users.forms.is_user_admin',
               return_value=as_admin), \
            patch('plinth.modules.users.views.is_user_admin',
                  return_value=as_admin), \
            patch('plinth.modules.users.views.update_session_auth_hash'):

        response = view(request, **kwargs)

    return response, messages


def test_users_list_view(rf):
    """Test users list view has correct view data."""
    with (patch('plinth.views.AppView.get_context_data',
                return_value={'is_enabled': True}),
          patch('plinth.views.AppView.app', return_value=None)):
        view = views.UserList.as_view()
        response, messages = make_request(rf.get('/'), view)

    assert response.status_code == 200


@pytest.mark.parametrize('username', ['test-new', 'test-neW@2_'])
def test_create_user_view(rf, username):
    """Test that user creation with valid usernames succeeds."""
    password = 'testingtesting'
    form_data = {
        'username': username,
        'password1': password,
        'password2': password,
        'confirm_password': 'adminpassword',
    }

    request = rf.post(urls.reverse('users:create'), data=form_data)
    view = views.UserCreate.as_view()
    response, messages = make_request(request, view)

    assert list(messages)[0].message == 'User {} created.'.format(username)
    assert response.status_code == 302
    assert response.url == urls.reverse('users:index')


@pytest.mark.parametrize('password,error', [
    ('', {
        'confirm_password': ['This field is required.']
    }),
    ('wrong_password', {
        'confirm_password': ['Invalid password.']
    }),
])
def test_create_user_invalid_admin_view(rf, password, error):
    """Test that user creation with an invalid admin password fails."""
    user_password = 'testingtesting'
    form_data = {
        'username': 'test-new',
        'password1': user_password,
        'password2': user_password,
        'confirm_password': password,
    }

    request = rf.post(urls.reverse('users:create'), data=form_data)
    view = views.UserCreate.as_view()
    response, messages = make_request(request, view)

    assert response.context_data['form'].errors == error


def test_create_user_form_view(rf):
    """Test that a username field on create form has correct attributes."""
    request = rf.get(urls.reverse('users:create'))
    view = views.UserCreate.as_view()
    response, messages = make_request(request, view)

    assert response.context_data['form'].fields['username'].widget.attrs == {
        'maxlength': '150',
        'autofocus': 'autofocus',
        'autocapitalize': 'none',
        'autocomplete': 'username'
    }
    assert response.status_code == 200


@pytest.mark.parametrize('username',
                         ['-tester', 'testér', 'test+test', 't', 'test test'])
def test_create_user_invalid_username_view(rf, username):
    """Test that user creation with invalid username fails."""
    form_data = {
        'username': username,
        'password1': 'testingtesting',
        'password2': 'testingtesting',
        'confirm_password': 'adminpassword',
    }

    request = rf.post(urls.reverse('users:create'), data=form_data)
    view = views.UserCreate.as_view()
    response, messages = make_request(request, view)

    assert response.context_data['form'].errors['username'][
        0] == 'Enter a valid username.'
    assert response.status_code == 200


@pytest.mark.parametrize(
    'username,error',
    [('testuser', 'A user with that username already exists.'),
     ('root', 'Username is taken or is reserved.'),
     ('rooT', 'Username is taken or is reserved.'),
     ('debian-minetest', 'Username is taken or is reserved.')])
def test_create_user_taken_or_reserved_username_view(rf, username, error):
    """Test that user creation with taken or reserved username fails."""
    User.objects.create(username='testuser')
    form_data = {
        'username': username,
        'password1': 'testingtesting',
        'password2': 'testingtesting',
        'confirm_password': 'adminpassword',
    }

    request = rf.post(urls.reverse('users:create'), data=form_data)
    view = views.UserCreate.as_view()
    response, messages = make_request(request, view)

    assert response.context_data['form'].errors['username'][0] == error
    assert response.status_code == 200


def test_update_user_view(rf):
    """Test that updating username succeeds."""
    user = 'tester'
    new_username = 'tester-renamed'
    form_data = {
        'username': new_username,
        'password1': 'testingtesting',
        'password2': 'testingtesting',
        'confirm_password': 'adminpassword',
    }

    url = urls.reverse('users:edit', kwargs={'slug': user})
    request = rf.post(url, data=form_data)
    view = views.UserUpdate.as_view()
    response, messages = make_request(request, view, as_admin=True, slug=user)

    assert list(messages)[0].message == 'User {} updated.'.format(new_username)
    assert response.status_code == 302
    assert response.url == urls.reverse('users:edit',
                                        kwargs={'slug': new_username})


@pytest.mark.parametrize('password,error', [
    ('', {
        'confirm_password': ['This field is required.']
    }),
    ('wrong_password', {
        'confirm_password': ['Invalid password.']
    }),
])
def test_update_user_invalid_admin_view(rf, password, error):
    """Test that updating username with an invalid admin password fails."""
    user = 'tester'
    new_username = 'tester-renamed'
    form_data = {
        'username': new_username,
        'password1': 'testingtesting',
        'password2': 'testingtesting',
        'confirm_password': password,
    }

    url = urls.reverse('users:edit', kwargs={'slug': user})
    request = rf.post(url, data=form_data)
    view = views.UserUpdate.as_view()
    response, messages = make_request(request, view, as_admin=True, slug=user)

    assert response.context_data['form'].errors == error


def test_update_user_without_permissions_view(rf):
    """Test that updating other user as non-admin user raises exception."""
    form_data = {
        'username': 'admin-renamed',
        'password1': 'testingtesting',
        'password2': 'testingtesting',
        'confirm_password': 'adminpassword',
    }

    url = urls.reverse('users:edit', kwargs={'slug': 'admin'})
    request = rf.post(url, data=form_data)
    view = views.UserUpdate.as_view()

    with pytest.raises(PermissionDenied):
        make_request(request, view, as_admin=False, slug='admin')


def test_delete_user_view(rf):
    """Test that user deletion succeeds."""
    user = 'tester'
    form_data = {
        'username': user,
        'delete': True,
        'confirm_password': 'adminpassword',
    }

    url = urls.reverse('users:edit', kwargs={'slug': user})
    request = rf.post(url, data=form_data)
    view = views.UserUpdate.as_view()
    response, messages = make_request(request, view, as_admin=True, slug=user)

    assert list(messages)[0].message == 'User {} deleted.'.format(user)
    assert response.status_code == 302
    assert response.url == urls.reverse('users:index')


def test_user_change_password_view(rf):
    """Test that changing password succeeds."""
    user = 'admin'
    form_data = {
        'new_password1': 'testingtesting2',
        'new_password2': 'testingtesting2',
        'confirm_password': 'adminpassword',
    }

    url = urls.reverse('users:change_password', kwargs={'slug': user})
    request = rf.post(url, data=form_data)
    view = views.UserChangePassword.as_view()
    response, messages = make_request(request, view, as_admin=True, slug=user)

    assert list(messages)[0].message == 'Password changed successfully.'
    assert response.status_code == 302
    assert response.url == urls.reverse('users:edit', kwargs={'slug': user})


@pytest.mark.parametrize('password,error', [
    ('', {
        'confirm_password': ['This field is required.']
    }),
    ('wrong_password', {
        'confirm_password': ['Invalid password.']
    }),
])
def test_user_change_password_invalid_admin_view(rf, password, error):
    """Test that changing password with an invalid admin password fails."""
    user = 'admin'
    form_data = {
        'new_password1': 'testingtesting2',
        'new_password2': 'testingtesting2',
        'confirm_password': password,
    }

    url = urls.reverse('users:change_password', kwargs={'slug': user})
    request = rf.post(url, data=form_data)
    view = views.UserChangePassword.as_view()
    response, messages = make_request(request, view, as_admin=True, slug=user)

    assert response.context_data['form'].errors == error
    assert response.status_code == 200


def test_user_change_password_without_permissions_view(rf):
    """
    Test that changing other user password as a non-admin user raises
    exception.
    """
    user = 'admin'
    form_data = {
        'new_password1': 'adminadmin2',
        'new_password2': 'adminadmin2',
        'confirm_password': 'adminpassword',
    }

    url = urls.reverse('users:change_password', kwargs={'slug': user})
    request = rf.post(url, data=form_data)
    view = views.UserChangePassword.as_view()

    with pytest.raises(PermissionDenied):
        make_request(request, view, as_admin=False, slug=user)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for backups.
"""

from django.utils.translation import gettext_lazy as _

# Currently, backup application does not have any settings. However, settings
# such as scheduler settings, backup location, secrets to connect to remove
# servers need to be backed up.
backup: dict = {}

tags = [
    _('Restore'),
    _('Encrypted'),
    _('Schedules'),
    _('Local'),
    _('Remote'),
    _('App data'),
    _('Configuration'),
    _('Borg')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app to configure system date and time.
"""

import subprocess

from django.utils.translation import gettext_lazy as _
from django.utils.translation import gettext_noop

from plinth import app as app_module
from plinth import menu
from plinth.daemon import Daemon, RelatedDaemon
from plinth.diagnostic_check import DiagnosticCheck, Result
from plinth.modules.backups.components import BackupRestore
from plinth.package import Packages

from . import manifest

_description = [
    _('Network time server is a program that maintains the system time '
      'in synchronization with servers on the Internet.')
]


class DateTimeApp(app_module.App):
    """FreedomBox app for date and time if time syncronization is unmanaged."""

    app_id = 'datetime'

    _version = 2

    _time_managed = None

    @property
    def can_be_disabled(self):
        """Return whether the app can be disabled."""
        return self._is_time_managed()

    def _is_time_managed(self):
        """Check whether time should be syncronized by the systemd-timesyncd.

        systemd-timesyncd does not run if we have another NTP daemon installed
        or FreedomBox runs inside a container where the host manages the time.

        """
        if self._time_managed is None:
            try:
                # Replace with the command 'systemd-analyze condition
                # --unit=systemd-timesyncd.service' when --unit argument
                # becomes available in a newer systemd.
                process = subprocess.run(
                    ['systemd-detect-virt', '--container'], check=False,
                    stdout=subprocess.PIPE)
                self._time_managed = (
                    process.stdout.decode().strip() == 'none')
            except (FileNotFoundError, subprocess.CalledProcessError):
                # When systemd-timesyncd will not run.
                self._time_managed = False

        return self._time_managed

    def __init__(self) -> None:
        """Create components for the app."""
        super().__init__()

        info = app_module.Info(app_id=self.app_id, version=self._version,
                               is_essential=True, name=_('Date & Time'),
                               icon='fa-clock-o', description=_description,
                               manual_page='DateTime', tags=manifest.tags)
        self.add(info)

        menu_item = menu.Menu('menu-datetime', info.name, info.icon, info.tags,
                              'datetime:index',
                              parent_url_name='system:system', order=40)
        self.add(menu_item)

        packages = Packages('packages-datetime', ['systemd-timesyncd'])
        self.add(packages)

        related_daemon = RelatedDaemon('daemon-datetime-timedated',
                                       'systemd-timedated')
        self.add(related_daemon)

        if self._is_time_managed():
            daemon = Daemon('daemon-datetime', 'systemd-timesyncd')
            self.add(daemon)

        backup_restore = BackupRestore('backup-restore-datetime',
                                       **manifest.backup)
        self.add(backup_restore)

    def diagnose(self) -> list[DiagnosticCheck]:
        """Run diagnostics and return the results."""
        results = super().diagnose()
        if self._is_time_managed():
            results.append(_diagnose_time_synchronized())

        return results

    def has_diagnostics(self):
        """Return that app has diagnostics only when time is managed."""
        return self._is_time_managed()

    def setup(self, old_version):
        """Install and configure the app."""
        super().setup(old_version)
        self.enable()


def _diagnose_time_synchronized() -> DiagnosticCheck:
    """Check whether time is synchronized to NTP server."""
    result = Result.FAILED
    try:
        output = subprocess.check_output(
            ['timedatectl', 'show', '--property=NTPSynchronized', '--value'])
        if 'yes' in output.decode():
            result = Result.PASSED
    except subprocess.CalledProcessError:
        pass

    return DiagnosticCheck('datetime-ntp-sync',
                           gettext_noop('Time synchronized to NTP server'),
                           result)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox app for configuring Shadowsocks Server."""

import random
import string

from django.contrib import messages
from django.utils.translation import gettext_lazy as _

from plinth import views

from . import privileged
from .forms import ShadowsocksServerForm


class ShadowsocksServerAppView(views.AppView):
    """Configuration view for Shadowsocks Server."""

    app_id = 'shadowsocksserver'
    form_class = ShadowsocksServerForm

    def get_initial(self, *args, **kwargs):
        """Get initial values for form."""
        status = super().get_initial()
        try:
            status.update(privileged.get_config())
        except Exception:
            # If we cannot read the configuration for some reason, generate a
            # new random password.
            password = ''.join(
                random.choice(string.ascii_letters) for _ in range(12))
            status.update({
                'password': password,
                'method': 'chacha20-ietf-poly1305',
            })

        return status

    def form_valid(self, form):
        """Configure Shadowsocks Server."""
        old_status = form.initial
        new_status = form.cleaned_data

        if old_status['password'] != new_status['password'] or \
           old_status['method'] != new_status['method']:
            privileged.merge_config(new_status['password'],
                                    new_status['method'])
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Syncthing."""

import os
import shutil
import time

import augeas

from plinth import action_utils
from plinth.actions import privileged

DATA_DIR = '/var/lib/syncthing'
# legacy configuration file
CONF_FILE_LEGACY = DATA_DIR + '/.config/syncthing/config.xml'
# configuration file since Debian Trixie if '.config/syncthing' directory
# doesn't exist
CONF_FILE = DATA_DIR + '/.local/state/syncthing/config.xml'


def augeas_load(conf_file):
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.add_transform('Xml.lns', conf_file)
    aug.load()
    return aug


@privileged
def setup():
    """Perform post-install actions for Syncthing."""
    # Create a 'syncthing' system user and group, if needed.
    action_utils.run(['systemd-sysusers', 'freedombox-syncthing.conf'])

    if not os.path.exists(DATA_DIR):
        os.makedirs(DATA_DIR, mode=0o750)
        shutil.chown(DATA_DIR, user='syncthing', group='syncthing')


@privileged
def setup_config():
    """Make configuration changes."""
    # wait until the configuration file is created by the syncthing daemon
    conf_file_in_use = CONF_FILE
    timeout = 300
    while timeout > 0:
        if os.path.exists(CONF_FILE_LEGACY):
            conf_file_in_use = CONF_FILE_LEGACY
            break
        elif os.path.exists(CONF_FILE):
            break
        timeout = timeout - 1
        time.sleep(1)

    aug = augeas_load(conf_file_in_use)

    # disable authentication missing notification as FreedomBox itself
    # provides authentication
    # also make sure no authentication is required on top of FreedomBox's
    configs_to_remove = (
        'options/unackedNotificationID[#text="authenticationUserAndPassword"]',
        'gui/user/#text', 'gui/password/#text')
    conf_changed = False
    for config in configs_to_remove:
        removed = bool(
            aug.remove('/files' + conf_file_in_use +
                       f'/configuration/{config}'))
        if removed:
            conf_changed = True

    configs = {
        # disable usage reporting notification by declining reporting
        # if the user has not made a choice yet
        'options/urAccepted/#text': '-1',
        # Set all the values that, misconfigured from the Syncthing UI,
        # can make the UI inaccessible. Such misconfigurations can be
        # corrected if the user re-runs the setup.
        # https://discuss.freedombox.org/t/solved-cant-access-syncthing-administration-panel/2137
        'gui/#attribute/tls': 'false',
        'gui/#attribute/enabled': 'true',
        'gui/address/#text': '127.0.0.1:8384'
    }
    for key, value in configs.items():
        config = f'/configuration/{key}'
        if aug.get('/files' + conf_file_in_use + config) != value:
            aug.set('/files' + conf_file_in_use + config, value)
            conf_changed = True

    aug.save()

    if conf_changed:
        action_utils.service_try_restart('syncthing@syncthing')


@privileged
def uninstall():
    """Remove configuration directory when app is uninstalled."""
    # legacy location
    shutil.rmtree(DATA_DIR + '/.config', ignore_errors=True)
    shutil.rmtree(DATA_DIR + '/.local', ignore_errors=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Manage DNS entries needed for an email server.

See: https://en.wikipedia.org/wiki/MX_record
See: https://dmarcguide.globalcyberalliance.org/
See: https://support.google.com/a/answer/2466580
See: https://datatracker.ietf.org/doc/html/rfc6186
See: https://rspamd.com/doc/modules/dkim_signing.html
See: https://en.wikipedia.org/wiki/Reverse_DNS_lookup
"""

import ipaddress
import typing
from dataclasses import dataclass

from plinth.modules.privacy import lookup_public_address

from . import privileged


@dataclass
class Entry:  # pylint: disable=too-many-instance-attributes
    """A DNS entry."""

    type_: str
    value: str
    domain: str | None = None
    class_: str = 'IN'
    ttl: int = 60
    priority: int = 10
    weight: int | None = None
    port: int | None = None

    def get_split_value(self):
        """If the record is TXT and value > 255, split it."""
        if len(self.value) <= 255:
            return self.value

        pieces = []
        value = self.value
        while value:
            pieces.append(f'"{value[:255]}"')
            value = value[255:]

        return ' '.join(pieces)


def get_entries(domain: str) -> list[Entry]:
    """Return the list of DNS entries to be set in DNS server for domain."""
    mx_spam_entries = [
        Entry(type_='MX', value=f'{domain}.'),
        Entry(type_='TXT', value='v=spf1 mx a ~all'),
        Entry(
            domain='_dmarc', type_='TXT',
            value='v=DMARC1; p=none; sp=quarantine; '
            f'rua=mailto:postmaster@{domain}; ')
    ]
    try:
        dkim_public_key = privileged.get_dkim_public_key(domain)
        dkim_entries = [
            Entry(domain='dkim._domainkey', type_='TXT',
                  value=f'v=DKIM1; k=rsa; p={dkim_public_key}')
        ]
    except Exception:
        dkim_entries = []

    autoconfig_entries = [
        Entry(domain='_submission._tcp', type_='SRV', weight=10, port=587,
              value=f'{domain}.'),
        Entry(domain='_imaps._tcp', type_='SRV', weight=10, port=993,
              value=f'{domain}.'),
        Entry(domain='_pop3s._tcp', type_='SRV', priority=20, weight=10,
              port=995, value=f'{domain}.'),
    ]
    return mx_spam_entries + dkim_entries + autoconfig_entries


def get_reverse_entries(domain: str) -> list[Entry]:
    """Return the list of reverse DNS entries to make."""
    entries = []
    for ip_type in typing.get_args(typing.Literal['ipv4', 'ipv6']):
        try:
            ip_address = lookup_public_address(ip_type)
            reverse_pointer = ipaddress.ip_address(ip_address).reverse_pointer
        except Exception as exception:
            reverse_pointer = \
                f'Error querying external {ip_type} address: {exception}'

        entry = Entry(domain=reverse_pointer, type_='PTR', value=f'{domain}.')
        entries.append(entry)

    return entries



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Django views for Deluge."""

from django.contrib import messages
from django.utils.translation import gettext as _

from plinth import views

from . import privileged
from .forms import DelugeForm


class DelugeAppView(views.AppView):
    """Serve configuration page."""

    diagnostics_module_name = 'deluge'
    form_class = DelugeForm
    app_id = 'deluge'

    def get_initial(self):
        """Get current Deluge server settings."""
        status = super().get_initial()
        configuration = privileged.get_configuration()
        status['storage_path'] = configuration['download_location']
        return status

    def form_valid(self, form):
        """Apply the changes submitted in the form."""
        old_status = form.initial
        new_status = form.cleaned_data

        if old_status['storage_path'] != new_status['storage_path']:
            privileged.set_configuration(
                download_location=new_status['storage_path'])
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure Matrix-Synapse server."""

import hashlib
import hmac
import json
import os
import pathlib
import shutil

import requests
import yaml

from plinth import action_utils, utils
from plinth.actions import privileged

CONF_DIR = "/etc/matrix-synapse/conf.d/"

ORIG_CONF_PATH = '/etc/matrix-synapse/homeserver.yaml'
SERVER_NAME_PATH = CONF_DIR + 'server_name.yaml'
STATIC_CONF_PATH = CONF_DIR + 'freedombox-static.yaml'
LISTENERS_CONF_PATH = CONF_DIR + 'freedombox-listeners.yaml'
REGISTRATION_CONF_PATH = CONF_DIR + 'freedombox-registration.yaml'
REGISTRATION_SECRET_PATH = CONF_DIR + 'freedombox-registration-secret.yaml'
ADMIN_ACCESS_TOKEN_PATH = CONF_DIR + 'freedombox-admin-access-token.txt'
TURN_CONF_PATH = CONF_DIR + 'freedombox-turn.yaml'
OVERRIDDEN_TURN_CONF_PATH = CONF_DIR + 'turn.yaml'
FREEDOMBOX_ADMIN_USERNAME = 'freedombox-admin'
ADMIN_API_BASE = 'http://localhost:8008/_synapse/admin/v1/'

STATIC_CONFIG = {
    'max_upload_size': '100M',
    'password_providers': [{
        'module': 'ldap_auth_provider.LdapAuthProvider',
        'config': {
            'enabled': True,
            'uri': 'ldap://localhost:389',
            'start_tls': False,
            'base': 'ou=users,dc=thisbox',
            'attributes': {
                'uid': 'uid',
                'name': 'uid',
                'mail': '',
            },
        },
    }, ],
    'suppress_key_server_warning': True,
    'trusted_key_servers': [{
        'server_name': 'matrix.org'
    }, ],
}


@privileged
def post_install():
    """Perform post installation configuration."""
    with open(STATIC_CONF_PATH, 'w', encoding='utf-8') as static_conf_file:
        yaml.dump(STATIC_CONFIG, static_conf_file)

    # start with listener config from original homeserver.yaml
    with open(ORIG_CONF_PATH, encoding='utf-8') as orig_conf_file:
        orig_config = yaml.safe_load(orig_conf_file)

    listeners = orig_config['listeners']
    for listener in listeners:
        if listener['port'] == 8448:
            listener['bind_addresses'] = ['::', '0.0.0.0']
            listener.pop('bind_address', None)

    with open(LISTENERS_CONF_PATH, 'w',
              encoding='utf-8') as listeners_conf_file:
        yaml.dump({'listeners': listeners}, listeners_conf_file)


@privileged
def setup(domain_name: str):
    """Configure the domain name for matrix-synapse package."""
    action_utils.dpkg_reconfigure('matrix-synapse',
                                  {'server-name': domain_name})


def get_config():
    """Return the current configuration of matrix-synapse."""
    try:
        with open(REGISTRATION_CONF_PATH, encoding='utf-8') as reg_conf_file:
            config = yaml.safe_load(reg_conf_file)
    except FileNotFoundError:
        # Check if its set in original conffile.
        with open(ORIG_CONF_PATH, encoding='utf-8') as orig_conf_file:
            config = yaml.safe_load(orig_conf_file)

    if config.get('enable_registration_without_verification'):
        registration_verification = 'disabled'
    elif config.get('registration_requires_token'):
        registration_verification = 'token'
    else:
        registration_verification = None

    return {
        'public_registration': bool(config.get('enable_registration', False)),
        'registration_verification': registration_verification,
    }


@privileged
def set_config(public_registration: bool,
               registration_verification: str | None = None):
    """Enable/disable public user registration."""
    if registration_verification == 'token':
        _create_registration_token()

    config = {'enable_registration': public_registration}
    if public_registration and registration_verification in ('disabled', None):
        config['enable_registration_without_verification'] = True
    elif registration_verification == 'token':
        config['registration_requires_token'] = True

    with open(REGISTRATION_CONF_PATH, 'w', encoding='utf-8') as reg_conf_file:
        yaml.dump(config, reg_conf_file)

    action_utils.service_try_restart('matrix-synapse')


@privileged
def move_old_conf():
    """Move old configuration to backup so it can be restored by reinstall."""
    conf_file = pathlib.Path(ORIG_CONF_PATH)
    if conf_file.exists():
        backup_file = conf_file.with_suffix(conf_file.suffix + '.fbx-bak')
        conf_file.replace(backup_file)


def _set_turn_config(conf_file, conf):
    turn_server_config = json.loads(conf)

    if not turn_server_config['uris']:
        # No valid configuration, remove the configuration file
        try:
            os.remove(conf_file)
        except FileNotFoundError:
            pass

        return

    config = {
        'turn_uris': turn_server_config['uris'],
        'turn_shared_secret': turn_server_config['shared_secret'],
        'turn_user_lifetime': 86400000,
        'turn_allow_guests': True
    }

    with open(conf_file, 'w+', encoding='utf-8') as turn_config:
        yaml.dump(config, turn_config)


@privileged
def configure_turn(managed: bool, conf: str):
    """Set parameters for the STUN/TURN server to use with Matrix Synapse."""
    if managed:
        _set_turn_config(TURN_CONF_PATH, conf)
    else:
        _set_turn_config(OVERRIDDEN_TURN_CONF_PATH, conf)

    action_utils.service_try_restart('matrix-synapse')


@privileged
def fix_public_registrations():
    """If public registrations are enabled, set validation mechanism."""
    config = get_config()
    if (config['public_registration']
            and config['registration_verification'] is None):
        set_config(public_registration=True,
                   registration_verification='disabled')


def _ensure_server_is_running():
    """Raise an exception if matrix-synapse server is not running."""
    if not action_utils.service_is_running('matrix-synapse'):
        raise ProcessLookupError


def _get_registration_shared_secret() -> str:
    """Ensure that server is set to use a shared secret for registration."""
    secret_file = pathlib.Path(REGISTRATION_SECRET_PATH)
    if secret_file.exists():
        with secret_file.open() as file_handle:
            return yaml.safe_load(file_handle)['registration_shared_secret']

    secret = utils.generate_password()
    # Ensure that file is readable by matrix-synapse user and root.
    previous_umask = os.umask(0o077)
    try:
        secret_file.write_text(f'registration_shared_secret: "{secret}"\n')
    finally:
        os.umask(previous_umask)

    shutil.chown(secret_file, 'matrix-synapse', 'nogroup')
    action_utils.service_try_restart('matrix-synapse')
    return secret


def _parse_api_error(response):
    """Raise exception if the response of API call was an error."""
    if response.get('errcode'):
        raise ConnectionError(response['errcode'], response['error'])


def _get_access_token() -> str:
    """Create a new freedombox-admin account and return its access token.

    For user registration Admin API, see:
    https://matrix-org.github.io/synapse/latest/admin_api/register_api.html
    """
    access_token_path = pathlib.Path(ADMIN_ACCESS_TOKEN_PATH)
    if access_token_path.exists():
        return access_token_path.read_text().strip()

    shared_secret = _get_registration_shared_secret()
    nonce = _get_nonce()
    username = FREEDOMBOX_ADMIN_USERNAME
    # No need to store password, we will use the access token.
    password = utils.generate_password()
    mac = _generate_mac(shared_secret, nonce, username, password, True)
    data = {
        'nonce': nonce,
        'username': username,
        'displayname': 'FreedomBox Admin',
        'password': password,
        'admin': True,
        'mac': mac
    }
    request = requests.post(ADMIN_API_BASE + 'register', json=data)
    response = request.json()
    _parse_api_error(response)

    # Ensure that file is only readable by root user.
    previous_umask = os.umask(0o077)
    try:
        access_token_path.write_text(response['access_token'])
    finally:
        os.umask(previous_umask)

    return response['access_token']


@privileged
def list_registration_tokens() -> list[dict[str, str | int | None]]:
    """Return the current list of registration tokens."""
    if not action_utils.service_is_running('matrix-synapse'):
        return []

    access_token = _get_access_token()
    return _list_registration_tokens(access_token)


def _get_headers(access_token: str):
    """Return the common HTTP headers needed for synapse admin API.

    For details on authorization to the Admin API, see:
    https://matrix-org.github.io/synapse/latest/usage/administration/admin_api/index.html
    """
    return {'Authorization': f'Bearer {access_token}'}


def _list_registration_tokens(
        access_token: str) -> list[dict[str, str | int | None]]:
    """Use Admin API to fetch the list of registration tokens.

    For details on registration tokens API, see:
    https://matrix-org.github.io/synapse/latest/usage/administration/admin_api/registration_tokens.html
    """
    request = requests.get(ADMIN_API_BASE + 'registration_tokens',
                           headers=_get_headers(access_token))
    response = request.json()
    _parse_api_error(response)
    return response['registration_tokens']


def _create_registration_token():
    """Make sure that at least one registration token is created.

    For details on registration tokens API, see:
    https://matrix-org.github.io/synapse/latest/usage/administration/admin_api/registration_tokens.html
    """
    _ensure_server_is_running()

    access_token = _get_access_token()
    tokens = _list_registration_tokens(access_token)
    for token in tokens:
        if token['uses_allowed'] is None and token['expiry_time'] is None:
            # A token with unlimited uses and for unlimited time already
            # exists.
            return

    if not tokens:
        # API can generate a token for us but it includes special chars which
        # is awkward to deal with for users.
        token = utils.generate_password(size=12)
        request = requests.post(ADMIN_API_BASE + 'registration_tokens/new',
                                headers=_get_headers(access_token),
                                json={'token': token})
        response = request.json()
        _parse_api_error(response)


def _get_nonce() -> str:
    """Make a preliminary registration request to get a nonce.

    Nonce must be returned to the server during actual registration request. It
    is used in computation of MAC and prevents replay attacks.
    https://matrix-org.github.io/synapse/latest/admin_api/register_api.html
    """
    request = requests.get(ADMIN_API_BASE + 'register')
    response = request.json()
    _parse_api_error(response)
    return response['nonce']


def _generate_mac(shared_secret: str, nonce: str, user: str, password: str,
                  admin: bool = False, user_type=None) -> str:
    """Generate MAC using HMAC-SHA1 algorithm.

    For information on how to encode the data for MAC computation, see:
    https://matrix-org.github.io/synapse/latest/admin_api/register_api.html
    """
    mac = hmac.new(
        key=shared_secret.encode('utf8'),
        digestmod=hashlib.sha1,
    )

    mac.update(nonce.encode('utf8'))
    mac.update(b'\x00')
    mac.update(user.encode('utf8'))
    mac.update(b'\x00')
    mac.update(password.encode('utf8'))
    mac.update(b'\x00')
    mac.update(b'admin' if admin else b'notadmin')
    if user_type:
        mac.update(b'\x00')
        mac.update(user_type.encode('utf8'))

    return mac.hexdigest()


@privileged
def uninstall():
    """Delete configuration and data directories."""
    for item in ['/etc/matrix-synapse/conf.d', '/var/lib/matrix-synapse']:
        shutil.rmtree(item, ignore_errors=True)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configure miscellaneous system settings."""

import os
import pathlib

import augeas

from plinth import action_utils
from plinth.actions import privileged

APACHE_CONF_ENABLED_DIR = '/etc/apache2/conf-enabled'
APACHE_HOMEPAGE_CONF_FILE_NAME = 'freedombox-apache-homepage.conf'
APACHE_HOMEPAGE_CONFIG = os.path.join(APACHE_CONF_ENABLED_DIR,
                                      APACHE_HOMEPAGE_CONF_FILE_NAME)

JOURNALD_FILE = pathlib.Path('/etc/systemd/journald.conf.d/50-freedombox.conf')


def load_augeas():
    """Initialize Augeas."""
    aug = augeas.Augeas(flags=augeas.Augeas.NO_LOAD +
                        augeas.Augeas.NO_MODL_AUTOLOAD)
    aug.transform('Puppet', str(JOURNALD_FILE))
    aug.set('/augeas/context', '/files' + str(JOURNALD_FILE))
    aug.load()
    return aug


def get_logging_mode() -> str:
    """Return the logging mode as none, volatile or persistent."""
    aug = load_augeas()
    storage = aug.get('Journal/Storage')
    if storage in ('volatile', 'persistent', 'none'):
        return storage

    # journald's default is 'auto'. On Debian systems, 'auto' is same
    # 'persistent' because /var/log/journal exists by default.
    return 'persistent'


@privileged
def set_logging_mode(mode: str):
    """Set the current logging mode."""
    if mode not in ('volatile', 'persistent', 'none'):
        raise ValueError('Invalid mode')

    aug = load_augeas()
    aug.set('Journal/Storage', mode)
    if mode == 'volatile':
        aug.set('Journal/MaxFileSec', '6h')
        aug.set('Journal/MaxRetentionSec', '2day')
    else:
        aug.remove('Journal/MaxFileSec')
        aug.remove('Journal/MaxRetentionSec')

    aug.remove('Journal/RuntimeMaxUse')
    JOURNALD_FILE.parent.mkdir(exist_ok=True)
    aug.save()

    # systemd-journald is socket activated, it may not be running and it does
    # not support reload.
    action_utils.service_try_restart('systemd-journald')


@privileged
def set_home_page(homepage: str | None):
    """Set the default app for this FreedomBox."""
    conf_file_path = os.path.join('/etc/apache2/conf-available',
                                  APACHE_HOMEPAGE_CONF_FILE_NAME)

    redirect_rule = ''
    if homepage:
        redirect_rule = 'RedirectMatch "^/$" "{}"\n'.format(homepage)

    with open(conf_file_path, 'w', encoding='utf-8') as conf_file:
        conf_file.write(redirect_rule)

    action_utils.webserver_enable('freedombox-apache-homepage')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""FreedomBox views for basic system configuration."""

from django.contrib import messages
from django.utils.translation import gettext as _

from plinth import views
from plinth.modules import config

from . import privileged
from .forms import ConfigurationForm


class ConfigAppView(views.AppView):
    """Serve configuration page."""

    form_class = ConfigurationForm
    app_id = 'config'

    def get_initial(self):
        """Return the current status."""
        return {
            'homepage': config.get_home_page(),
            'advanced_mode': config.get_advanced_mode(),
            'logging_mode': privileged.get_logging_mode(),
        }

    def form_valid(self, form):
        """Apply the form changes."""
        old_status = form.initial
        new_status = form.cleaned_data

        is_changed = False

        if old_status['homepage'] != new_status['homepage']:
            try:
                config.change_home_page(new_status['homepage'])
            except Exception as exception:
                messages.error(
                    self.request,
                    _('Error setting webserver home page: {exception}').format(
                        exception=exception))
            else:
                messages.success(self.request, _('Webserver home page set'))

        if old_status['advanced_mode'] != new_status['advanced_mode']:
            try:
                config.set_advanced_mode(new_status['advanced_mode'])
            except Exception as exception:
                messages.error(
                    self.request,
                    _('Error changing advanced mode: {exception}').format(
                        exception=exception))
            else:
                if new_status['advanced_mode']:
                    messages.success(self.request,
                                     _('Showing advanced apps and features'))
                else:
                    messages.success(self.request,
                                     _('Hiding advanced apps and features'))

        if old_status['logging_mode'] != new_status['logging_mode']:
            privileged.set_logging_mode(new_status['logging_mode'])
            is_changed = True

        if is_changed:
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for names module.
"""

import pytest

from .. import on_domain_added, on_domain_removed
from ..components import DomainName, DomainType


@pytest.fixture(name='clean_domains')
def fixture_clean_domains():
    """Test fixture to start a test with clean domains list."""
    DomainName._all = {}  # pylint: disable=protected-access


@pytest.mark.usefixtures('clean_domains')
def test_on_domain_added():
    """Test adding a domain to the global list."""
    on_domain_added('', '')
    assert not DomainName.list()

    DomainType('domain-type-tor', 'Tor Domain', 'torurl')
    on_domain_added('tor', 'domain-type-tor', 'ddddd.onion')
    on_domain_added('tor', 'domain-type-tor', 'eeeee.onion')
    DomainName.get('domain-tor-ddddd.onion')
    DomainName.get('domain-tor-eeeee.onion')


@pytest.mark.usefixtures('clean_domains')
def test_on_domain_removed():
    """Test removing a domain from the global list."""
    DomainType('domain-type-tor', 'Tor Domain', 'torurl')
    on_domain_added('tor', 'domain-type-tor', 'ddddd.onion')
    on_domain_removed('tor', 'domain-type-tor', 'ddddd.onion')
    with pytest.raises(KeyError):
        DomainName.get('domain-tor-ddddd.onion')

    # try to remove things that don't exist
    on_domain_removed('', '')
    with pytest.raises(KeyError):
        on_domain_removed('', 'some-domain-type', 'x-unknown-name')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for Shadowsocks Server.
"""

from django.utils.translation import gettext_lazy as _

backup = {
    'secrets': {
        'files': [
            '/var/lib/private/shadowsocks-libev/fbxserver/fbxserver.json'
        ]
    },
    'services': ['shadowsocks-libev-server@fbxserver']
}

tags = [
    _('Censorship resistance'),
    _('Encrypted tunnel'),
    _('Exit point'),
    _('Shadowsocks')
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Application manifest for cockpit.
"""

from django.utils.translation import gettext_lazy as _

clients = [{
    'name': _('Cockpit'),
    'platforms': [{
        'type': 'web',
        'url': '/_cockpit/'
    }]
}]

# cockpit.conf need not be backed up because add/remove domain signals are
# triggered on every Plinth domain change (and cockpit application install) and
# will set the value of allowed domains correctly. This is the only key the is
# customized in cockpit.conf.
backup: dict = {}

tags = [
    _('Advanced administration'),
    _('Web terminal'),
    _('Storage'),
    _('Networking'),
    _('Services'),
    _('Logs'),
    _('Performance'),
]



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Functional, browser based tests for wireguard app.
"""

import urllib

import pytest

from plinth.tests import functional

pytestmark = [pytest.mark.apps, pytest.mark.wireguard]


class TestWireguardApp(functional.BaseAppTests):
    app_name = 'wireguard'
    has_web = False

    _client_public_key1 = '8dlHfuAsU0WsnpQkNd8fguwmBtFX5/CkO0xYKbNBt24='
    _client_public_key2 = 'xM/vyXbYtMf9bDTZY3W2x7Moxw4m1ckpnZvVSd4DLho='

    _server = [
        {
            'peer_endpoint': 'wg1.example.org:1234',
            'peer_public_key': 'HBCqZk4B93N6q19zNleJkAVs+PEfWAPgPpKnrhL/CVw=',
            'ip_address_and_network': '10.0.0.2/32',
            'private_key': '',
            'preshared_key': ''
        },
        {
            'peer_endpoint': 'wg2.example.org:5678',
            'peer_public_key': 'Z/iHo0vaeSN8Ykk5KwhQ819MMU5nyzD7y7xFFthlxXI=',
            'ip_address_and_network': '192.168.0.2/24',
            'private_key': 'QC2xEZMn3bgNsSVFrU51+ALSUiUaWg6gRWigh3EeVm0=',
            'preshared_key': 'AHxZ4Rr8Ij4L1aq+ceusSIgBfluqiI9Vb5I2UtQFanI='
        },
    ]

    def test_backup_restore(self, session_browser):
        """Skip the test."""
        pytest.skip('Does not implement backup/restore.')

    @staticmethod
    def _get_client_href(key):
        """Return the href for client show page."""
        key = urllib.parse.quote(urllib.parse.quote(key, safe=''))
        return f'/freedombox/apps/wireguard/client/{key}/show/'

    def _client_exists(self, browser, key):
        """Check whether a client key exists."""
        functional.nav_to_module(browser, 'wireguard')
        href = browser.links.find_by_href(self._get_client_href(key))
        return bool(href)

    def _assert_client_details(self, browser, key):
        """Check that the client's details are correct."""

        def get_value(key):
            return browser.find_by_css(f'tr.{key} > td').first.text

        functional.nav_to_module(browser, 'wireguard')
        href = browser.links.find_by_href(self._get_client_href(key))
        href.first.click()
        assert get_value('client-public-key') == key

    @staticmethod
    def _add_client(browser, key):
        """Add a client."""
        functional.nav_to_module(browser, 'wireguard')
        # Start the server on FreedomBox, if needed.
        start_server_button = browser.find_by_css('.btn-start-server')
        if start_server_button:
            with functional.wait_for_page_update(browser):
                start_server_button.first.click()

        browser.find_by_css('.btn-add-client').first.click()
        browser.find_by_id('id_public_key').fill(key)
        functional.submit(browser, form_class='form-add-client')

    def _edit_client(self, browser, key1, key2):
        """Edit a client"""
        functional.nav_to_module(browser, 'wireguard')
        browser.links.find_by_href(self._get_client_href(key1)).first.click()
        browser.find_by_css('.btn-edit-client').first.click()
        browser.find_by_id('id_public_key').fill(key2)
        functional.submit(browser, form_class='form-edit-client')

    def _delete_client(self, browser, key):
        """Delete a client"""
        functional.nav_to_module(browser, 'wireguard')
        browser.links.find_by_href(self._get_client_href(key)).first.click()
        browser.find_by_css('.btn-delete-client').first.click()
        functional.submit(browser, form_class='form-delete-client')

    def test_add_edit_delete_client(self, session_browser):
        """Test adding, editing and deleting a WireGuard client."""
        if self._client_exists(session_browser, self._client_public_key1):
            self._delete_client(session_browser, self._client_public_key1)

        if self._client_exists(session_browser, self._client_public_key2):
            self._delete_client(session_browser, self._client_public_key2)

        self._add_client(session_browser, self._client_public_key1)
        assert self._client_exists(session_browser, self._client_public_key1)

        self._edit_client(session_browser, self._client_public_key1,
                          self._client_public_key2)
        assert not self._client_exists(session_browser,
                                       self._client_public_key1)
        assert self._client_exists(session_browser, self._client_public_key2)

        self._delete_client(session_browser, self._client_public_key2)
        assert not self._client_exists(session_browser,
                                       self._client_public_key2)

    def test_auto_add_client(self, session_browser):
        """Test the automatic client generation and addition flow."""
        functional.nav_to_module(session_browser, 'wireguard')

        # Start server if needed (reuse existing logic)
        # Extract to reusable method
        start_server_button = session_browser.find_by_css('.btn-start-server')
        if start_server_button:
            with functional.wait_for_page_update(session_browser):
                start_server_button.first.click()

        session_browser.find_by_css('.btn-auto-add-client').first.click()

        client_pubkey = session_browser.find_by_css(
                '.pubkey-val').first.text.strip()

        # Verify private key reveal
        privkey_reveal = session_browser.find_by_css(
                '.privkey-val')
        assert privkey_reveal, "Private key reveal should be present"
        privkey_reveal.click()
        client_privkey = session_browser.find_by_css(
                '.privkey-val').text.splitlines()[1]
        assert len(client_privkey) == 44, (("Private key should be base64 "
                                            "(44 chars)"))

        # Verify config download and QR links
        download_link = session_browser.links.find_by_href(
                '/freedombox/apps/wireguard/client/auto-add/action/download/')
        qr_link = session_browser.links.find_by_href(
                '/freedombox/apps/wireguard/client/auto-add/action/qr/')
        assert download_link, "Download config link should exist"
        assert qr_link, "QR code link should exist"

        # Submit to add the client
        with functional.wait_for_page_update(session_browser):
            session_browser.find_by_css(
                    '.btn-auto-add-connection').first.click()

        # Verify client was added successfully
        assert self._client_exists(session_browser, client_pubkey), ((
            "Auto-generated client should exist"))

        # Clean up
        self._delete_client(session_browser, client_pubkey)
        assert not self._client_exists(session_browser, client_pubkey)

    @staticmethod
    def _get_server_href(browser, key):
        """Return the href for server show page."""
        return browser.find_by_css(
            f'.peer-public-key[data-public-key="{key}"] a')

    def _server_exists(self, browser, config):
        """Check whether a server key exists."""
        functional.nav_to_module(browser, 'wireguard')
        href = self._get_server_href(browser, config['peer_public_key'])
        return bool(href)

    def _assert_server_details(self, browser, config):
        """Check that the server's details are correct."""

        def get_value(key):
            return browser.find_by_css(f'tr.{key} > td').first.text

        functional.nav_to_module(browser, 'wireguard')
        href = self._get_server_href(browser, config['peer_public_key'])
        href.first.click()
        assert get_value('peer-endpoint') == config['peer_endpoint']
        assert get_value('peer-public-key') == config['peer_public_key']
        assert get_value('server-ip-address-and-network'
                         ) == config['ip_address_and_network']
        assert get_value('peer-preshared-key') == (config['preshared_key']
                                                   or 'None')

    @staticmethod
    def _add_server(browser, config):
        """Add a server."""
        functional.nav_to_module(browser, 'wireguard')
        browser.find_by_css('.btn-add-server').first.click()
        browser.find_by_id('id_peer_endpoint').fill(config['peer_endpoint'])
        browser.find_by_id('id_peer_public_key').fill(
            config['peer_public_key'])
        browser.find_by_id('id_ip_address_and_network').fill(
            config['ip_address_and_network'])
        browser.find_by_id('id_private_key').fill(config['private_key'])
        browser.find_by_id('id_preshared_key').fill(config['preshared_key'])
        functional.submit(browser, form_class='form-add-server')

    def _edit_server(self, browser, config1, config2):
        """Edit a server."""
        functional.nav_to_module(browser, 'wireguard')
        self._get_server_href(browser,
                              config1['peer_public_key']).first.click()
        browser.find_by_css('.btn-edit-server').first.click()
        browser.find_by_id('id_peer_endpoint').fill(config2['peer_endpoint'])
        browser.find_by_id('id_peer_public_key').fill(
            config2['peer_public_key'])
        browser.find_by_id('id_ip_address_and_network').fill(
            config2['ip_address_and_network'])
        browser.find_by_id('id_private_key').fill(config2['private_key'])
        browser.find_by_id('id_preshared_key').fill(config2['preshared_key'])
        functional.submit(browser, form_class='form-edit-server')

    def _delete_server(self, browser, config):
        """Delete a server"""
        functional.nav_to_module(browser, 'wireguard')
        self._get_server_href(browser, config['peer_public_key']).first.click()
        browser.find_by_css('.btn-delete-server').first.click()
        functional.submit(browser, form_class='form-delete-server')

    def test_add_edit_delete_server(self, session_browser):
        """Test adding, editing and deleting a WireGuard server."""
        if self._server_exists(session_browser, self._server[0]):
            self._delete_server(session_browser, self._server[0])

        if self._server_exists(session_browser, self._server[1]):
            self._delete_server(session_browser, self._server[1])

        self._add_server(session_browser, self._server[0])
        self._assert_server_details(session_browser, self._server[0])

        # XXX: Editing the connection is very flaky. Fails, incorrect and
        # crashes. Fix it.

        # self._edit_server(session_browser, self._server[0],
        # self._server[1]) assert not self._server_exists(session_browser,
        # self._server[0]) self._assert_server_details(session_browser,
        # self._server[1])

        # self._delete_server(session_browser, self._server[1])
        # assert not self._server_exists(session_browser, self._server[1])

        self._delete_server(session_browser, self._server[0])
        assert not self._server_exists(session_browser, self._server[0])



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Configuration helper for the ejabberd service."""

import logging
import os
import re
import shutil
import socket
import subprocess
from pathlib import Path
from typing import Any

from ruamel.yaml import YAML, scalarstring

from plinth import action_utils
from plinth.actions import privileged
from plinth.version import Version

logger = logging.getLogger(__name__)

EJABBERD_CONFIG = '/etc/ejabberd/ejabberd.yml'
EJABBERD_BACKUP = '/var/log/ejabberd/ejabberd.dump'
EJABBERD_BACKUP_NEW = '/var/log/ejabberd/ejabberd_new.dump'
EJABBERD_ORIG_CERT = '/etc/ejabberd/ejabberd.pem'
EJABBERD_MANAGED_COTURN = '/etc/ejabberd/freedombox_managed_coturn'
IQDISC_DEPRECATED_VERSION = Version('18.03')
MOD_IRC_DEPRECATED_VERSION = Version('18.06')

yaml = YAML()
yaml.allow_duplicate_keys = True
yaml.preserve_quotes = True  # type: ignore [assignment]

TURN_URI_REGEX = r'(stun|turn):(.*):([0-9]{4})(?:\?transport=(tcp|udp))?'


@privileged
def pre_install(domain_name: str):
    """Preseed debconf values before packages are installed."""
    if not domain_name:
        # If new domain_name is blank, use hostname instead.
        domain_name = socket.gethostname()

    action_utils.debconf_set_selections(
        ['ejabberd ejabberd/hostname string ' + domain_name])


@privileged
def setup() -> None:
    """Enable LDAP authentication."""
    with open(EJABBERD_CONFIG, 'r', encoding='utf-8') as file_handle:
        conf = yaml.load(file_handle)

    for listen_port in conf['listen']:
        if 'tls' in listen_port:
            listen_port['tls'] = False
        if 'use_turn' in listen_port:
            conf['listen'].remove(listen_port)  # Use coturn instead
        if listen_port['port'] == 5443:
            # Enable XEP-0363 HTTP File Upload
            listen_port['request_handlers']['/upload'] = 'mod_http_upload'

    origin_key = scalarstring.DoubleQuotedScalarString(
        'Access-Control-Allow-Origin')
    origin_value = scalarstring.DoubleQuotedScalarString('https://@HOST@')
    methods_key = scalarstring.DoubleQuotedScalarString(
        'Access-Control-Allow-Methods')
    methods_value = scalarstring.DoubleQuotedScalarString(
        'GET,HEAD,PUT,OPTIONS')
    headers_key = scalarstring.DoubleQuotedScalarString(
        'Access-Control-Allow-Headers')
    headers_value = scalarstring.DoubleQuotedScalarString('Content-Type')
    conf['modules']['mod_http_upload'] = {
        'put_url': 'https://@HOST@/upload',
        'custom_headers': {
            origin_key: origin_value,
            methods_key: methods_value,
            headers_key: headers_value,
        },
    }

    conf['auth_method'] = 'ldap'
    conf['ldap_servers'] = [scalarstring.DoubleQuotedScalarString('localhost')]
    conf['ldap_base'] = scalarstring.DoubleQuotedScalarString(
        'ou=users,dc=thisbox')

    # Read all available certificates
    conf['certfiles'].append(
        scalarstring.DoubleQuotedScalarString(
            '/etc/ejabberd/letsencrypt/*/ejabberd.pem'))
    conf['certfiles'] = list(set(conf['certfiles']))

    with open(EJABBERD_CONFIG, 'w', encoding='utf-8') as file_handle:
        yaml.dump(conf, file_handle)

    _upgrade_config()

    try:
        action_utils.run(['ejabberdctl', 'restart'], check=True)
    except subprocess.CalledProcessError as err:
        logger.warn('Failed to restart ejabberd with new configuration: %s',
                    err)


def _upgrade_config() -> None:
    """Fix the config file by removing deprecated settings."""
    current_version = _get_version()
    if not current_version:
        logger.warn('Warning: Unable to get ejabberd version.')

    with open(EJABBERD_CONFIG, 'r', encoding='utf-8') as file_handle:
        conf = yaml.load(file_handle)

    # Check if `iqdisc` is present and remove it
    if 'mod_mam' in conf['modules'] and \
       (not current_version or current_version > IQDISC_DEPRECATED_VERSION):
        conf['modules']['mod_mam'].pop('iqdisc', None)

    # check if mod_irc is present in modules and remove it
    if 'mod_irc' in conf['modules'] and \
       (not current_version or current_version > MOD_IRC_DEPRECATED_VERSION):
        conf['modules'].pop('mod_irc')

    # Debian has a patch to configuration to change port 5443 to 5280 in
    # ejabberd package version 18.12. However, 5443 is the correct port to host
    # BOSH. So, change it back. In 19.x, this behavior has changed to use both
    # ports 5443 (for BOSH) and 5280 (for web administration).
    bosh_port = any((True for listen_port in conf['listen']
                     if listen_port['port'] == 5443))
    if not bosh_port:
        for listen_port in conf['listen']:
            if listen_port['port'] == 5280:
                listen_port['port'] = 5443

    # s2s_certfile is deprecated in favor of certfiles
    if 's2s_certfile' in conf:
        conf.pop('s2s_certfile')

    # Write changes back to the file
    with open(EJABBERD_CONFIG, 'w', encoding='utf-8') as file_handle:
        yaml.dump(conf, file_handle)


@privileged
def pre_change_hostname(old_hostname: str, new_hostname: str):
    """Prepare ejabberd for hostname change."""
    if not shutil.which('ejabberdctl'):
        logger.info('ejabberdctl not found')
        return

    action_utils.run(['ejabberdctl', 'backup', EJABBERD_BACKUP], check=False)
    action_utils.run([
        'ejabberdctl', 'mnesia-change-nodename', 'ejabberd@' + old_hostname,
        'ejabberd@' + new_hostname, EJABBERD_BACKUP, EJABBERD_BACKUP_NEW
    ], check=True)
    os.remove(EJABBERD_BACKUP)


@privileged
def change_hostname():
    """Update ejabberd with new hostname."""
    if not shutil.which('ejabberdctl'):
        return

    action_utils.service_stop('ejabberd')
    action_utils.run(['pkill', '-u', 'ejabberd'], check=False)

    # Make sure there aren't files in the Mnesia spool dir
    os.makedirs('/var/lib/ejabberd/oldfiles', exist_ok=True)
    action_utils.run('mv /var/lib/ejabberd/*.* /var/lib/ejabberd/oldfiles/',
                     shell=True, check=False)

    action_utils.service_start('ejabberd')

    # restore backup database
    if os.path.exists(EJABBERD_BACKUP_NEW):
        try:
            action_utils.run(['ejabberdctl', 'restore', EJABBERD_BACKUP_NEW],
                             check=True)
            os.remove(EJABBERD_BACKUP_NEW)
        except subprocess.CalledProcessError as err:
            logger.error('Failed to restore ejabberd backup database: %s', err)
    else:
        logger.error('Could not load ejabberd backup database: %s not found' %
                     EJABBERD_BACKUP_NEW)


@privileged
def get_domains() -> list[str]:
    """Get all configured domains."""
    if not shutil.which('ejabberdctl'):
        return []

    with open(EJABBERD_CONFIG, 'r', encoding='utf-8') as file_handle:
        conf = yaml.load(file_handle)

    return conf['hosts']


@privileged
def add_domain(domain_name: str):
    """Update ejabberd with new domain name.

    Restarting ejabberd is handled by letsencrypt-ejabberd component.
    """
    if not shutil.which('ejabberdctl'):
        logger.info('ejabberdctl not found')
        return

    # Add updated domain name to ejabberd hosts list.
    with open(EJABBERD_CONFIG, 'r', encoding='utf-8') as file_handle:
        conf = yaml.load(file_handle)

    conf['hosts'].append(scalarstring.DoubleQuotedScalarString(domain_name))

    conf['hosts'] = list(set(conf['hosts']))

    with open(EJABBERD_CONFIG, 'w', encoding='utf-8') as file_handle:
        yaml.dump(conf, file_handle)

    # Restarting ejabberd is handled by letsencrypt-ejabberd component.


@privileged
def set_domains(domains: list[str]):
    """Set list of ejabberd domains.

    Restarting ejabberd is handled by letsencrypt-ejabberd component.
    """
    if not len(domains):
        raise ValueError('No domains provided')

    if not shutil.which('ejabberdctl'):
        return

    with open(EJABBERD_CONFIG, 'r', encoding='utf-8') as file_handle:
        conf = yaml.load(file_handle)

    conf['hosts'] = domains

    with open(EJABBERD_CONFIG, 'w', encoding='utf-8') as file_handle:
        yaml.dump(conf, file_handle)


@privileged
def mam(command: str) -> bool | None:
    """Enable, disable, or get status of Message Archive Management (MAM)."""
    if command not in ('enable', 'disable', 'status'):
        raise ValueError('Invalid command')

    with open(EJABBERD_CONFIG, 'r', encoding='utf-8') as file_handle:
        conf = yaml.load(file_handle)

    if 'modules' not in conf:
        return None

    if command == 'status':
        return 'mod_mam' in conf['modules']

    if command == 'enable':
        # Explicitly set the recommended / default settings for mod_mam,
        # see https://docs.ejabberd.im/admin/configuration/#mod-mam.
        settings_mod_mam = {
            'mod_mam': {
                'db_type':
                    'mnesia',  # default is 'mnesia' (w/o set default_db)
                'default': 'always',  # helps various clients to use mam
                'request_activates_archiving': False,  # default False
                'assume_mam_usage': False,  # for non-ack'd msgs, default False
                'cache_size': 1000,  # default is 1000 items
                'cache_life_time': 3600  # default is 3600 seconds = 1h
            }
        }
        conf['modules'].update(settings_mod_mam)
    elif command == 'disable':
        # disable modules by erasing from config file
        if 'mod_mam' in conf['modules']:
            conf['modules'].pop('mod_mam')

    with open(EJABBERD_CONFIG, 'w', encoding='utf-8') as file_handle:
        yaml.dump(conf, file_handle)

    if action_utils.service_is_running('ejabberd'):
        action_utils.run(['ejabberdctl', 'reload_config'], check=False)

    return None


def _generate_service(uri: str) -> dict:
    """Generate ejabberd mod_stun_disco service config from Coturn URI."""
    pattern = re.compile(TURN_URI_REGEX)
    match = pattern.match(uri)
    if not match:
        raise ValueError('URL does not match TURN URI')

    typ, domain, port, transport = match.groups('udp')
    return {
        "host": domain,
        "port": int(port),
        "type": typ,
        "transport": transport,
    }


def _generate_uris(services: list[dict]) -> list[str]:
    """Generate STUN/TURN URIs from ejabberd mod_stun_disco service config."""
    uris = []
    for s in services:
        uri = f"{s['type']}:{s['host']}:{s['port']}"
        if s['type'] != 'stun':
            uri += f"?transport={s['transport']}"

        if uri not in uris:
            uris.append(uri)

    return uris


@privileged
def get_turn_config() -> tuple[dict[str, Any], bool]:
    """Get the latest STUN/TURN configuration in JSON format."""
    with open(EJABBERD_CONFIG, 'r', encoding='utf-8') as file_handle:
        conf = yaml.load(file_handle)

    mod_stun_disco_config = conf['modules']['mod_stun_disco']
    managed = os.path.exists(EJABBERD_MANAGED_COTURN)

    if bool(mod_stun_disco_config):
        return {
            'domain': '',
            'uris': _generate_uris(mod_stun_disco_config['services']),
            'shared_secret': mod_stun_disco_config['secret'],
        }, managed
    else:
        return {'domain': None, 'uris': [], 'shared_secret': None}, managed


@privileged
def configure_turn(turn_server_config: dict[str, Any], managed: bool):
    """Set parameters for the STUN/TURN server to use with ejabberd."""
    uris = turn_server_config['uris']
    mod_stun_disco_config = {}

    if turn_server_config['uris'] and turn_server_config['shared_secret']:
        mod_stun_disco_config = {
            'credentials_lifetime': '1000d',
            'secret': turn_server_config['shared_secret'],
            'services': [_generate_service(uri) for uri in uris]
        }

    with open(EJABBERD_CONFIG, 'r', encoding='utf-8') as file_handle:
        conf = yaml.load(file_handle)

    conf['modules']['mod_stun_disco'] = mod_stun_disco_config

    with open(EJABBERD_CONFIG, 'w', encoding='utf-8') as file_handle:
        yaml.dump(conf, file_handle)

    if managed:
        Path(EJABBERD_MANAGED_COTURN).touch()
    else:
        Path(EJABBERD_MANAGED_COTURN).unlink(missing_ok=True)

    if action_utils.service_is_running('ejabberd'):
        action_utils.run(['ejabberdctl', 'reload_config'], check=False)


def _get_version():
    """Get the current ejabberd version."""
    try:
        output = action_utils.run(['ejabberdctl', 'status'],
                                  check=True).stdout.decode('utf-8')
    except subprocess.CalledProcessError:
        return None

    version_info = output.strip().split('\n')[-1].split()
    if version_info:
        version = str(version_info[1])
        return Version(version)
    return None



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Views for privacy app."""

from django.contrib import messages
from django.utils.translation import gettext as _

import plinth.modules.names.privileged as names_privileged
from plinth.modules import names, privacy
from plinth.modules.privacy.forms import PrivacyForm
from plinth.views import AppView

from . import privileged


class PrivacyAppView(AppView):
    """Serve configuration page."""

    app_id = 'privacy'
    form_class = PrivacyForm

    def get_initial(self):
        """Return the values to fill in the form."""
        initial = super().get_initial()
        initial.update(privileged.get_configuration())
        initial['ip_lookup_url'] = privacy.get_ip_lookup_url()
        if names.is_resolved_installed():
            initial.update(names_privileged.get_resolved_configuration())

        return initial

    def form_valid(self, form):
        """Change the configurations of Minetest service."""
        new_config = form.cleaned_data
        old_config = form.initial

        changes = {}
        is_changed = False
        if old_config['enable_popcon'] != new_config['enable_popcon']:
            changes['enable_popcon'] = new_config['enable_popcon']

        if 'dns_fallback' in old_config:
            if old_config['dns_fallback'] != new_config['dns_fallback']:
                names_privileged.set_resolved_configuration(
                    dns_fallback=new_config['dns_fallback'])
                is_changed = True

        if old_config['ip_lookup_url'] != new_config['ip_lookup_url']:
            privacy.set_ip_lookup_url(new_config['ip_lookup_url'])
            is_changed = True

        if changes:
            privileged.set_configuration(**changes)

        if changes or is_changed:
            messages.success(self.request, _('Configuration updated'))

        return super().form_valid(form)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Test component to manage a container using podman."""

from unittest.mock import call, patch

import pytest

from plinth.app import App, Info
from plinth.container import Container
from plinth.diagnostic_check import DiagnosticCheck, Result

pytestmark = pytest.mark.usefixtures('mock_privileged')
privileged_modules_to_mock = [
    'plinth.privileged', 'plinth.privileged.container',
    'plinth.privileged.service'
]


class AppTest(App):
    """Test application that contains a daemon."""

    app_id = 'test-app'


@pytest.fixture(name='container')
def fixture_container():
    app1 = AppTest()
    app1.add(Info('test-app', 1))
    container = Container('test-container', 'name1', 'image:stable', 'volume1',
                          '/volume', {'/host1': '/cont1'}, {'KEY1': 'VAL1'},
                          ['service1.service'], {'/dev/host1': '/dev/cont1'},
                          [(1234, 'tcp4')])
    app1.add(container)
    with patch('plinth.app.App.list') as app_list:
        app_list.return_value = [app1]
        yield container


def test_container_init(container):
    """Test initializing the container component."""
    component = Container('test-container', 'name1', 'image:stable', 'volume1',
                          '/volume')
    assert component.component_id == 'test-container'
    assert component.name == 'name1'
    assert component.image_name == 'image:stable'
    assert component.volume_name == 'volume1'
    assert component.volume_path == '/volume'
    assert component.volumes is None
    assert component.env is None
    assert component.binds_to is None
    assert component.devices is None
    assert component.listen_ports == []

    assert container.component_id == 'test-container'
    assert container.name == 'name1'
    assert container.image_name == 'image:stable'
    assert container.volume_name == 'volume1'
    assert container.volume_path == '/volume'
    assert container.volumes == {'/host1': '/cont1'}
    assert container.env == {'KEY1': 'VAL1'}
    assert container.binds_to == ['service1.service']
    assert container.devices == {'/dev/host1': '/dev/cont1'}
    assert container.listen_ports == [(1234, 'tcp4')]


@patch('plinth.action_utils.podman_is_enabled')
def test_container_is_enabled(podman_is_enabled, container):
    """Test checking if container is enabled."""
    podman_is_enabled.return_value = False
    assert not container.is_enabled()

    podman_is_enabled.return_value = True
    assert container.is_enabled()


@patch('plinth.action_utils.service_enable')
@patch('plinth.action_utils.podman_enable')
def test_container_enable(podman_enable, enable, container):
    """Test enabling a container component."""
    container.enable()
    assert podman_enable.mock_calls == [call('name1')]
    assert enable.mock_calls == [call('name1')]


@patch('plinth.action_utils.service_disable')
@patch('plinth.action_utils.podman_disable')
def test_container_disable(podman_disable, disable, container):
    """Test disabling a container component."""
    container.disable()
    assert podman_disable.mock_calls == [call('name1')]
    assert disable.mock_calls == [call('name1')]


@patch('plinth.action_utils.service_is_running')
def test_container_is_running(service_is_running, container):
    """Test checking of container component is running."""
    service_is_running.return_value = False
    assert not container.is_running()
    assert service_is_running.mock_calls == [call('name1')]

    service_is_running.reset_mock()
    service_is_running.return_value = True
    assert container.is_running()


@patch('plinth.action_utils.service_disable')
@patch('plinth.action_utils.service_enable')
@patch('plinth.action_utils.service_is_running')
def test_container_ensure_running(service_is_running, enable, disable,
                                  container):
    """Test checking of container component can be ensured to be running."""
    service_is_running.return_value = True
    with container.ensure_running() as state:
        assert state
        assert enable.mock_calls == []

    assert disable.mock_calls == []

    service_is_running.return_value = False
    with container.ensure_running() as state:
        assert not state
        assert enable.mock_calls == [call('name1')]

    assert disable.mock_calls == [call('name1')]


@patch('plinth.action_utils.service_disable')
@patch('plinth.action_utils.service_start')
@patch('plinth.action_utils.podman_disable')
@patch('plinth.action_utils.podman_is_enabled')
@patch('plinth.action_utils.podman_create')
def test_container_setup(podman_create, is_enabled, disable, service_start,
                         service_disable, container):
    """Test setting up the container."""
    is_enabled.return_value = True
    container.setup(0)
    assert podman_create.mock_calls == [
        call('name1', 'image:stable', 'volume1', '/volume',
             {'/host1': '/cont1'}, {'KEY1': 'VAL1'}, ['service1.service'],
             {'/dev/host1': '/dev/cont1'})
    ]
    assert service_start.mock_calls == [call('name1', check=True)]
    assert disable.mock_calls == []

    is_enabled.return_value = False
    container.setup(0)
    assert disable.mock_calls == []

    is_enabled.return_value = False
    container.setup(1)
    assert disable.mock_calls == [call('name1')]
    assert service_disable.mock_calls == [call('name1')]


@patch('plinth.action_utils.podman_uninstall')
def test_container_uninstall(podman_uninstall, container):
    """Test uninstalling the container."""
    container.uninstall()
    assert podman_uninstall.mock_calls == [
        call(container_name='name1', image_name='image:stable',
             volume_name='volume1', volume_path='/volume')
    ]


@patch('plinth.action_utils.service_is_running')
@patch('plinth.container.diagnose_port_listening')
def test_container_diagnose(diagnose_port_listening, service_is_running,
                            container):
    """Test diagnosing the container."""
    expected_results = [
        DiagnosticCheck('container-running-name1',
                        'Container {container_name} is running', Result.PASSED,
                        {'container_name': 'name1'}, 'test-container'),
        DiagnosticCheck('daemon-listening-tcp4-1234',
                        'Listening on tcp4 port 1234', Result.PASSED, {
                            'kind': 'tcp4',
                            'port': 1234
                        }, 'test-container'),
    ]
    diagnose_port_listening.return_value = expected_results[1]
    service_is_running.return_value = True
    results = container.diagnose()
    assert results == expected_results

    service_is_running.return_value = False
    expected_results[0].result = Result.FAILED
    results = container.diagnose()
    assert results == expected_results



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
GnuDIP client for updating Dynamic DNS records.
"""

import hashlib
import logging
import socket
from html.parser import HTMLParser

import requests

logger = logging.getLogger(__name__)


class MetaTagParser(HTMLParser):
    """Extracts name and content from HTML meta tags as a dictionary."""

    def __init__(self) -> None:
        """Initialize the meta tags."""
        super().__init__()
        self.meta: dict[str, str] = {}

    def handle_starttag(self, tag: str,
                        attrs: list[tuple[str, str | None]]) -> None:
        """Handle encountering an opening HTML tag during parsing."""
        if tag.lower() == 'meta':
            attr_dict = dict(attrs)
            name = attr_dict.get('name')
            content = attr_dict.get('content')
            if name and content:
                self.meta[name] = content


def _extract_content_from_meta_tags(html: str) -> dict[str, str]:
    """Return a dict of {name: content} for all meta tags in the HTML."""
    parser = MetaTagParser()
    parser.feed(html)
    return parser.meta


def _check_required_keys(dictionary: dict[str, str], keys: list[str]) -> None:
    missing_keys = [key for key in keys if key not in dictionary]
    if missing_keys:
        raise ValueError(
            f"Missing required keys in response: {', '.join(missing_keys)}")


def _request_get_ipv4(*args, **kwargs):
    """Make a IPv4-only request.

    XXX: This monkey-patches socket.getaddrinfo which may causes issues when
    running multiple threads. With urllib3 >= 2.4 (Trixie has 2.3), it is
    possible to implement more cleanly. Use a session for requests library. In
    the session add custom adapter for https:. In the adapter, override
    creation of pool manager, and pass socket_family parameter.
    """
    original = socket.getaddrinfo

    def getaddrinfo_ipv4(*args, **kwargs):
        return original(args[0], args[1], socket.AF_INET, *args[3:], **kwargs)

    socket.getaddrinfo = getaddrinfo_ipv4
    try:
        return requests.get(*args, **kwargs)
    finally:
        socket.getaddrinfo = original


def update(server: str, domain: str, username: str,
           password: str) -> tuple[bool, str | None]:
    """Update Dynamic DNS record using GnuDIP protocol.

    Protocol documentation:
    https://gnudip2.sourceforge.net/gnudip-www/latest/gnudip/html/protocol.html

    GnuDIP at least as deployed on the FreedomBox foundation servers does not
    support IPv6 (it does have any code to update AAAA records). So, make a
    request only using IPv4 stack.
    """
    domain = domain.removeprefix(username + '.')
    password_digest = hashlib.md5(password.encode()).hexdigest()

    http_server = f'https://{server}/gnudip/cgi-bin/gdipupdt.cgi'
    response = _request_get_ipv4(http_server)

    salt_response = _extract_content_from_meta_tags(response.text)
    _check_required_keys(salt_response, ['salt', 'time', 'sign'])

    salt = salt_response['salt']
    password_digest = hashlib.md5(
        f'{password_digest}.{salt}'.encode()).hexdigest()

    query_params = {
        'salt': salt,
        'time': salt_response['time'],
        'sign': salt_response['sign'],
        'user': username,
        'domn': domain,
        'pass': password_digest,
        'reqc': '2'
    }
    update_response = _request_get_ipv4(http_server, params=query_params)

    update_result = _extract_content_from_meta_tags(update_response.text)
    _check_required_keys(update_result, ['retc'])
    result = (int(update_result['retc']) == 0)
    return result, update_result.get('addr')



# SPDX-License-Identifier: AGPL-3.0-or-later

import os
from urllib.parse import urlparse

from django import template
from django.utils.safestring import mark_safe

from plinth import clients as clients_module
from plinth import utils, web_server

register = template.Library()


def mark_active_menuitem(menu, path):
    """Mark the best-matching menu item with 'active'

    Input: a menu dict in the form of:
        [{'url': '/path/to/choice1/', 'text': 'choice 1'}, {'url': ...}]

    URL paths are expected to end with a slash for matches to work properly.

    Output: The same dictionary; the best-matching URL dict gets the value
    'active': True. All other URL dicts get the value 'active': False.

    Note: this sets the 'active' values on the menu itself, not on a copy.
    """
    best_match = ''
    best_match_item = None

    for urlitem in menu:
        urlitem['active'] = False
        if not path.startswith(str(urlitem['url'])):
            continue

        match = os.path.commonprefix([str(urlitem['url']), path])
        if len(match) > len(best_match):
            best_match = match
            best_match_item = urlitem

    if best_match_item:
        best_match_item['active'] = True

    return menu


@register.inclusion_tag('subsubmenu.html', takes_context=True)
def show_subsubmenu(context, menu):
    """Mark the active menu item and display the subsubmenu"""
    menu = mark_active_menuitem(menu, context['request'].path)
    return {'subsubmenu': menu}


@register.filter(name='clients_of_type')
def clients_of_type(clients, client_type):
    """Filter and get clients of a particular type"""
    return clients_module.of_type(clients, client_type)


@register.filter(name='lookup')
def lookup(dictionary, key):
    """Get the value in the dictionary at given key"""
    return dictionary[key]


def _is_internal_url(url):
    """Check if the given link is internal or not.

    A URL is internal if it is relative URL or points to one of the domains
    managed by FreedomBox.
    """
    parsed_url = urlparse(str(url))
    if not parsed_url.netloc:
        return True

    from plinth.modules.names.components import DomainName
    return parsed_url.netloc in DomainName.list_names()


@register.filter(name='clients_get_platforms')
def clients_get_platforms(clients):
    """Return lists of self hosted platforms and all other platforms."""
    other = []
    web = []
    for client in clients:
        for platform in client['platforms']:
            if platform['type'] == 'web' and _is_internal_url(platform['url']):
                web.append(platform)
            else:
                other.append(platform)

    return {'web': web, 'other': other}


@register.simple_tag(name='icon')
def icon(url: str, *args, **kwargs):
    """Insert an SVG icon inline."""
    icon_name = url.rpartition('/')[2].partition('.')[0]

    def add_attributes(text: str) -> str:
        """Append attributes to the <svg> elment."""
        if 'class' not in kwargs:
            kwargs['class'] = 'svg-icon'

        kwargs['data-icon-name'] = icon_name
        attributes = ' '.join(
            (f'{key}="{value}"' for key, value in kwargs.items()))
        text = text.replace('<svg', f'<svg {attributes} ', count=1)
        text = text.replace('autoidmagic-', utils.random_string() + '-')
        return text

    if '/' not in url:
        # Only icon name specified
        url = f'theme/icons/{url}.svg'

    path = web_server.resolve_static_path(url)
    try:
        icon_text = path.read_text()
    except FileNotFoundError:
        raise ValueError(f'Icon {url} not found.')
    else:
        # Skip the line with <?xml> header.
        if icon_text and icon_text.startswith('<?xml'):
            icon_lines = icon_text.splitlines()
            icon_text = '\n'.join(icon_lines[1:])

        icon_text = add_attributes(icon_text)

    return mark_safe(icon_text)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
Tests for calibre views.
"""

from unittest.mock import call, patch

import pytest
from django import urls
from django.contrib.messages.storage.fallback import FallbackStorage
from django.http.response import Http404

from plinth import module_loader
from plinth.modules.calibre import views

# For all tests, use plinth.urls instead of urls configured for testing
pytestmark = pytest.mark.urls('plinth.urls')


@pytest.fixture(autouse=True, scope='module')
def fixture_calibre_urls():
    """Make sure calibre app's URLs are part of plinth.urls."""
    with patch('plinth.module_loader._modules_to_load', new=[]) as modules, \
            patch('plinth.urls.urlpatterns', new=[]):
        modules.append('plinth.modules.calibre')
        module_loader.include_urls()
        yield


@pytest.fixture(autouse=True)
def calibre_patch():
    """Patch calibre methods."""
    with patch('plinth.modules.calibre.privileged.list_libraries'
               ) as list_libraries:
        list_libraries.return_value = ['TestExistingLibrary']

        yield


def make_request(request, view, **kwargs):
    """Make request with a message storage."""
    setattr(request, 'session', 'session')
    messages = FallbackStorage(request)
    setattr(request, '_messages', messages)
    response = view(request, **kwargs)

    return response, messages


@patch('plinth.modules.calibre.privileged.create_library')
def test_create_library(create_library, rf):
    """Test that create library view works."""
    form_data = {'calibre-name': 'TestLibrary'}
    request = rf.post(urls.reverse('calibre:create-library'), data=form_data)
    view = views.CreateLibraryView.as_view()
    response, messages = make_request(request, view)

    assert response.status_code == 302
    assert response.url == urls.reverse('calibre:index')
    assert list(messages)[0].message == 'Library created.'
    create_library.assert_has_calls([call('TestLibrary')])


@patch('plinth.modules.calibre.privileged.create_library')
def test_create_library_failed(create_library, rf):
    """Test that create library fails as expected."""
    create_library.side_effect = RuntimeError('TestError')
    form_data = {'calibre-name': 'TestLibrary'}
    request = rf.post(urls.reverse('calibre:create-library'), data=form_data)
    view = views.CreateLibraryView.as_view()
    response, messages = make_request(request, view)

    assert response.status_code == 302
    assert response.url == urls.reverse('calibre:index')
    assert list(messages)[0].message == \
        'An error occurred while creating the library. TestError'


def test_create_library_existing_library(rf):
    """Test that create library errors out for an existing library name."""
    form_data = {'calibre-name': 'TestExistingLibrary'}
    request = rf.post(urls.reverse('calibre:create-library'), data=form_data)
    view = views.CreateLibraryView.as_view()
    response, _ = make_request(request, view)

    assert response.context_data['form'].errors['name'][
        0] == 'A library with this name already exists.'
    assert response.status_code == 200


def test_create_library_invalid_name(rf):
    """Test that create library errors out for invalid name."""
    form_data = {'calibre-name': 'Invalid Library'}
    request = rf.post(urls.reverse('calibre:create-library'), data=form_data)
    view = views.CreateLibraryView.as_view()
    response, _ = make_request(request, view)

    assert response.context_data['form'].errors['name'][
        0] == 'Enter a valid value.'
    assert response.status_code == 200


@patch('plinth.app.App.get')
def test_delete_library_confirmation_view(_app, rf):
    """Test that deleting library confirmation shows correct name."""
    response, _ = make_request(rf.get(''), views.delete_library,
                               name='TestExistingLibrary')
    assert response.status_code == 200
    assert response.context_data['name'] == 'TestExistingLibrary'


@patch('plinth.modules.calibre.privileged.delete_library')
@patch('plinth.app.App.get')
def test_delete_library(_app, delete_library, rf):
    """Test that deleting a library works."""
    response, messages = make_request(rf.post(''), views.delete_library,
                                      name='TestExistingLibrary')
    assert response.status_code == 302
    assert response.url == urls.reverse('calibre:index')
    assert list(messages)[0].message == 'TestExistingLibrary deleted.'
    delete_library.assert_has_calls([call('TestExistingLibrary')])


@patch('plinth.modules.calibre.privileged.delete_library')
def test_delete_library_error(delete_library, rf):
    """Test that deleting a library shows error when operation fails."""
    delete_library.side_effect = ValueError('TestError')
    response, messages = make_request(rf.post(''), views.delete_library,
                                      name='TestExistingLibrary')
    assert response.status_code == 302
    assert response.url == urls.reverse('calibre:index')
    assert list(messages)[0].message == \
        'Could not delete TestExistingLibrary: TestError'


def test_delete_library_non_existing(rf):
    """Test that deleting a library shows error when operation fails."""
    with pytest.raises(Http404):
        make_request(rf.post(''), views.delete_library,
                     name='TestNonExistingLibrary')

    with pytest.raises(Http404):
        make_request(rf.get(''), views.delete_library,
                     name='TestNonExistingLibrary')



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
FreedomBox app for running diagnostics.
"""

import logging

from django.contrib import messages
from django.http import Http404
from django.shortcuts import redirect
from django.template.response import TemplateResponse
from django.urls import NoReverseMatch, reverse
from django.utils.translation import gettext_lazy as _
from django.views.decorators.http import require_POST
from django.views.generic import TemplateView

from plinth import operation
from plinth.app import App
from plinth.diagnostic_check import Result
from plinth.modules import diagnostics
from plinth.setup import run_repair_on_app
from plinth.views import AppView

from .forms import ConfigureForm

logger = logging.getLogger(__name__)


class DiagnosticsView(AppView):
    """Diagnostics app page."""

    app_id = 'diagnostics'
    form_class = ConfigureForm
    template_name = 'diagnostics.html'

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        context = super().get_context_data(**kwargs)
        context['has_diagnostics'] = False
        context['results_available'] = diagnostics.are_results_available()
        return context

    def get_initial(self):
        """Return the initial values for the form."""
        status = super().get_initial()
        status['daily_run_enabled'] = diagnostics.is_daily_run_enabled()
        status['automatic_repair'] = diagnostics.is_automatic_repair_enabled()
        return status

    def form_valid(self, form):
        """Apply the form changes."""
        old_status = form.initial
        new_status = form.cleaned_data
        updated = False

        if old_status['daily_run_enabled'] != new_status['daily_run_enabled']:
            diagnostics.set_daily_run_enabled(new_status['daily_run_enabled'])
            updated = True

        if old_status['automatic_repair'] != new_status['automatic_repair']:
            diagnostics.set_automatic_repair_enabled(
                new_status['automatic_repair'])
            updated = True

        if updated:
            messages.success(self.request, _('Configuration updated.'))

        return super().form_valid(form)


class DiagnosticsFullView(TemplateView):
    """View to run full diagnostics."""

    template_name = 'diagnostics_full.html'

    def post(self, request):
        """Start diagnostics."""
        diagnostics.start_diagnostics()
        return self.get(self, request)

    def get_context_data(self, **kwargs):
        """Return additional context for rendering the template."""
        try:
            is_task_running = bool(operation.manager.get('diagnostics-full'))
        except KeyError:
            is_task_running = False

        context = super().get_context_data(**kwargs)
        context['is_task_running'] = is_task_running
        context['results'] = diagnostics.get_results()
        context['refresh_page_sec'] = 3 if is_task_running else None
        exception = context['results'].pop('exception', None)
        if exception:
            messages.error(self.request, exception)

        return context


@require_POST
def diagnose_app(request, app_id):
    """Return diagnostics for a particular app."""
    try:
        app = App.get(app_id)
    except KeyError:
        raise Http404('App does not exist')

    app_name = app.info.name or app_id
    diagnosis = app.diagnose()
    show_repair = False
    for check in diagnosis:
        if check.result in [Result.FAILED, Result.WARNING]:
            show_repair = True
            break

    return TemplateResponse(
        request, 'diagnostics_app.html', {
            'title': _('Diagnostic Test'),
            'app_id': app_id,
            'app_name': app_name,
            'results': diagnosis,
            'show_repair': show_repair,
        })


@require_POST
def repair(request, app_id):
    """Try to repair failed diagnostics on an app.

    Allows apps and components to customize the repair method. Re-run setup is
    the default if not specified.
    """
    try:
        app = App.get(app_id)
    except KeyError:
        raise Http404('App does not exist')

    try:
        finish_url = reverse(f'{app_id}:index')
    except NoReverseMatch:
        # for apps like apache that don't have an index route
        finish_url = reverse('diagnostics:index')

    current_version = app.get_setup_version()
    if not current_version:
        logger.warning('App %s is not installed, cannot repair', app_id)
        message = _('App {app_id} is not installed, cannot repair')
        messages.error(request, str(message).format(app_id=app_id))
        return redirect(finish_url)

    run_repair_on_app(app_id)
    return redirect(finish_url)



# SPDX-License-Identifier: AGPL-3.0-or-later
"""Manage application shortcuts on front page."""

import json
import logging
import pathlib
from typing import ClassVar

from plinth import app, cfg
from plinth.modules.users import privileged as users_privileged

logger = logging.getLogger(__name__)


class Shortcut(app.FollowerComponent):
    """An application component for handling shortcuts."""

    _all_shortcuts: ClassVar[dict[str, 'Shortcut']] = {}

    def __init__(self, component_id: str, name: str | None,
                 icon: str | None = None, url: str | None = None,
                 description: list[str] | None = None,
                 manual_page: str | None = None,
                 configure_url: str | None = None,
                 clients: list[dict] | None = None,
                 tags: list[str] | None = None, login_required: bool = False,
                 allowed_groups: list[str] | None = None):
        """Initialize the frontpage shortcut component for an app.

        When a user visits this web interface, they are first shown the
        frontpage. It's primary contents are the list of shortcuts to services
        that user may use on the server. If a service requires logging in, it
        does not show up to anonymous users. If a service requires a user to be
        part of a group, that service is only shown to those users.

        'component_id' must be a unique string across all apps and components
        of a app. Conventionally starts with 'shortcut-'.

        'name' is the mandatory title for the shortcut.

        'icon' is used to find a suitable image to represent the shortcut.

        'url' is link to which the user is redirected when the shortcut is
        activated. This is typically the web interface for a particular service
        provided by the app. For shortcuts that should simply additional
        information this value must be None.

        'description' is additional information that the user is shown when
        the shortcut is activated. This must be provided instead of 'url' and
        must be 'None' if 'url' is provided.

        'manual_page' is the optional name of the page for this app in the user
        manual. If provided, a 'Learn more...' link appears in the app page for
        this app.

        'configure_url' is the page to which the user may be redirected if they
        wish to change the settings for the app or one of its services. This is
        only used when 'url' is 'None'. It is optionally provided along with
        'description'.

        'clients' is a list of clients software that can used to access the
        service offered by the shortcut. This should be a valid client
        information structure as validated by clients.py:validate().

        'tags' is a list of tags that describe the app. Tags help users to find
        similar apps or alternatives and discover use cases.

        If 'login_required' is true, only logged-in users will be shown this
        shortcut. Anonymous users visiting the frontpage won't be shown this
        shortcut.

        'allowed_groups' specifies a list of user groups to whom this shortcut
        must be shown. All other user groups will not be shown this shortcut on
        the frontpage. If 'login_required' is False, this property has not
        effect and the shortcut is shown to all the users including anonymous
        users.

        """
        super().__init__(component_id)

        if url is None:
            url = '?selected={id}'.format(id=component_id)

        self.name = name
        self.url = url
        self.icon = icon
        self.description = description
        self.manual_page = manual_page
        self.configure_url = configure_url
        self.clients = clients
        self.tags = tags
        self.login_required = login_required
        self.allowed_groups = set(allowed_groups) if allowed_groups else None

        self._all_shortcuts[self.component_id] = self

    def remove(self):
        """Remove this shortcut from global list of shortcuts."""
        del self._all_shortcuts[self.component_id]

    @classmethod
    def list(cls, username=None, web_apps_only=False, sort_by='name'):
        """Return menu items in sorted order according to current locale."""
        shortcuts_to_return = cls._list_for_user(username)
        if web_apps_only:
            shortcuts_to_return = {
                _id: shortcut
                for _id, shortcut in shortcuts_to_return.items()
                if not shortcut.url.startswith('?selected=')
            }

        return sorted(shortcuts_to_return.values(),
                      key=lambda item: getattr(item, sort_by).lower())

    @classmethod
    def _list_for_user(cls, username=None):
        """Return menu items for a particular user or anonymous user."""
        if not username:
            return cls._all_shortcuts

        # XXX: Turn this into an API call in users module and cache
        user_groups = set(users_privileged.get_user_groups(username))

        if 'admin' in user_groups:  # Admin has access to all services
            return cls._all_shortcuts

        shortcuts = {}
        for shortcut_id, shortcut in cls._all_shortcuts.items():
            if shortcut.login_required and shortcut.allowed_groups and \
               user_groups.isdisjoint(shortcut.allowed_groups):
                continue

            shortcuts[shortcut_id] = shortcut

        return shortcuts


def add_custom_shortcuts():
    custom_shortcuts = get_custom_shortcuts()

    for shortcut in custom_shortcuts['shortcuts']:
        web_app_url = _extract_web_app_url(shortcut)
        if not web_app_url:
            continue

        shortcut_id = shortcut.get('id', shortcut['name'])
        component_id = 'shortcut-custom-' + shortcut_id
        tags = shortcut.get('tags', [])
        component = Shortcut(component_id, shortcut['name'],
                             icon=shortcut['icon_url'], tags=tags,
                             url=web_app_url)
        component.set_enabled(True)


def _extract_web_app_url(custom_shortcut):
    if not custom_shortcut.get('clients'):
        return None

    for client in custom_shortcut['clients']:
        if not client.get('platforms'):
            continue

        for platform in client['platforms']:
            if platform['type'] == 'web':
                return platform['url']

    return None


def get_custom_shortcuts_paths():
    """Return the list of custom shortcut file paths."""
    file_paths = [
        '/etc/freedombox/custom-shortcuts.json',
        '/etc/plinth/custom-shortcuts.json',
        '/var/lib/freedombox/custom-shortcuts.json',
        '/usr/share/freedombox/custom-shortcuts.json',
    ]
    return cfg.expand_to_dot_d_paths(file_paths)


def get_custom_shortcuts():
    """Return a merged dictionary of all custom shortcuts."""
    shortcuts = {'shortcuts': []}
    for file_path in get_custom_shortcuts_paths():
        file_path = pathlib.Path(file_path)
        try:
            if not file_path.is_file() or not file_path.stat().st_size:
                continue

            logger.info('Loading custom shortcuts from %s', file_path)
            with file_path.open(encoding='utf-8') as file_handle:
                shortcuts['shortcuts'] += json.load(file_handle)['shortcuts']
        except Exception as exception:
            logger.warning('Error loading shortcuts from %s: %s', file_path,
                           exception)

    return shortcuts



# SPDX-License-Identifier: AGPL-3.0-or-later
"""
URL for the Ejabberd module
"""

from django.urls import re_path

from .views import EjabberdAppView

urlpatterns = [
    re_path(r'^apps/ejabberd/$', EjabberdAppView.as_view(), name='index')
]
